/*
 *
 * tc902x.c
 *
 * Gigabit Ethernet device driver for Network Interface Cards
 * (NICs) utilizing the Sundance Technology Inc. TC902X
 * 1000Mbps Ethernet Media Access Controller
 *
 * Craig Rich
 * Sundance Technology, Inc.
 * 1485 Saratoga Avenue
 * Suite 200
 * San Jose, CA 95129
 * 408 873 4117
 * www.sundanceti.com
 * craig_rich@sundanceti.com
 *
 * Rev  Date     Description
 * --------------------------------------------------------------
 * 0.1  11/8/99  Initial revision work begins.
 *
 * 0.2  11/12/99  Basic operation achieved, continuing work.
 *
 * 0.3  11/19/99  MAC Loop Back for sync problem testing.
 *
 * 0.4  12/22/99  ioctl for diagnotic program 'hunter' support.
 *
 * 0.5  4/13/00   Updates to 
 *
 * 0.6  6/14/00   Slight correction to handling TFDDONE, and
 *                preservation of PHYCTRL polarity bits.
 *
 * 0.7  7/27/00   Modifications to accomodate triple speed
 *                autonegotiation. Also change to ioctl routine
 *                to handle unknown PHY address.
 *
 * 0.8  8/11/00   Added change_mtu function.
 *
 * 0.9  8/15/00   Corrected autonegotiation resolution.
 *
 * 0.10 8/30/00   Changed constants to use TC902X in place
 *                of RIO. Also, removed most of debug
 *                code in preparation for production release.
 *
 * 0.11 8/31/00   Utilize 64 bit data types where appropriate.
 *
 * 0.12 9/1/00    Move some constants to include file and utilize
 *                RxDMAInt register.
 *
 * 0.13 10/31/00  Several minor modifications to improve stability.
 *
 * 0.14 11/28/00  Added call to nic_tx_free if TFD not available.
 *
 * 0.15 12/5/00   Corrected problem with receive errors, always set
 *                receive buffer address to NULL. Release RX buffers
 *                on errors.
 *
 * 0.16 12/20/00  Corrected autoneg resolution issue, must detect
 *                speed via PHYCTRL register. Also, perform only 1
 *                loop in the nic_txcleanup routine.
 *
 * 0.17 2/7/01    Changed all references of ST2021 to TC902X.
 *                When next TFD not available, return -ENOMEM instead
 *                of 0. Removed references to RUBICON.
 *
 * 0.18 2/14/01   Corrected problem when unexpected jumbo frames are
 *                received (now dropped properly.) Changed
 *                "DROP_ON_ERRORS" breaking out Ethernet errors and
 *                TCP/IP errors serparately. Corrected Gigabit
 *                copper PAUSE autonegotiation.
 *
 * 0.19 2/22/01   Changed interrupt handling of RFD_LIST_END,
 *                INT_REQUESTED, and RX_DMA_COMPLETE. Masked off
 *                RMON statistics and unused MIB statistics.
 *                Make sure *all* statistics are accounted for
 *                (either masked or read in get_stats) to avoid
 *                perpetual UpdateStats interrupt from causing
 *                driver to crash.
 *
 * 0.20 3/2/01    Corrected error in nic_stop. Need to set
 *                TxBuff[] = NULL after freeing TxBuff and
 *                RxBuff[] = NULL after freeing RxBuff.
 *
 * 0.21 3/5/01    Correct 10/100Mbit PAUSE autonegotiation.
 *
 * 0.22 3/16/01   Used TxDMAIndicate for 100/1000Mbps modes. Add
 *                "TFD unavailable" and "RFD list end" counters
 *                to assist with performance measurement. Added
 *                check for maxtfdcnt != 0 to while loop within
 *                txcleanup.
 *
 * 0.23 3/22/01   Set the CurrentTxFrameID to 1 upon detecting
 *                a TxDMAComplete to reduce the number of TxDMAComplete.
 *                Also, indicate IP/TCP/UDP checksum is unneseccary
 *                if TC902X indicates checksum validates.
 *
 * 0.24 3/23/01   Changed the txfree routine, eliminating the margin
 *                between the last freed TFD and the current TFD.
 *
 * 0.25 4/3/01    Corrected errors in config_autoneg to deal with
 *                fiber flag properly.
 *
 * 0.26 5/1/01    Port for operation with Linux 2.2 or 2.4 kernel.
 *
 * 0.27 5/22/01   Cleaned up some extraneous comments.
 *
 * 0.28 6/20/01   Added auto IP, TCP, and UDP checksum addition
 *                on transmit based on compilation option.
 *
 * 0.29 7/26/01   Comment out #include <asm/spinlock.h> from tc902x.h
 *                for compatibility with RedHat 7.1. Unkown reason.
 *
 * 0.30 8/10/01   Added debug message to each function, print function
 *                name when entered. Added DEBUGCTRL register bit 5 for
 *                Rx DMA Poll Now bug work around. Added ifdef TC902X_DEBUG
 *                flags to TC902X_TFDlistunabvail and TX902X_RFDlistend
 *                counters. Removed clearing of sp->stat struct from
 *                nic_open and added check in get_stats to make sure
 *                NIC is initialized before reading statistic registers.
 *                Corrected erroneous MACCTRL setting for Fiber based
 *                10/100 boards. Corrected storage of phyctrlpolarity
 *                variable.
 *
 * 0.31 8/13/01   Incorporate STI or TMI fiber based NIC detection.
 *                Corrected problem with _pciremove_linux2_4 routine.
 *                Corrected setting of IP/TCP/UDP checksumming on receive
 *                and transmit.
 *
 * 0.32 8/15/01   Changed the tmi_fiber_detect routine.
 *
 * 0.33 8/16/01   Changed PHY reset method in nic_open routine. Added
 *                a chip reset in nic_stop to shut down the TC902X.
 *
 * 0.34 9/5/01    Corrected some misuage of dev_kfree_skb.
 *
 * 0.35 10/30/01  Unmap register space (IO or memory) in the nic_stop
 *                routine instead of in the cleanup or remove routines.
 *                Corrects driver up/down/up problem when using IO
 *                register mapping.
 *
 * 0.36 10/31/01  Modify the constant TC902X_FRAMESBETWEENTXDMACOMPLETES 
 *		  from 0x10 to 1. 
 */

#include "tc902x.h"

/* notify kernel that this software is under GNU Public License */
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10) )
    MODULE_LICENSE("GPL"); 
#endif
    
/* Function prototypes. */
u16     read_phy_register(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                          int phy_address, int phy_register);
int     tc902x_reset(u32 baseaddr, u32 resetflags);
int     tc902x_io_config(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
void    tc902x_interrupt_handler(int tc902x_irq, void *device_instance,
                                struct pt_regs *regs);

void    tc902x_nic_txcleanup(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
void	tc902x_nic_txfree(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int     tc902x_nic_open(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int     tc902x_nic_stop(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int     tc902x_nic_hard_start_xmit(struct sk_buff *skb,
                                   TC902X_DEVICE_TYPE *tc902x_ethernet_device);
TC902X_STATS_TYPE* tc902x_nic_get_stats(TC902X_DEVICE_TYPE
                                        *tc902x_ethernet_device);
void tc902x_nic_set_multicast_list(TC902X_DEVICE_TYPE *dev);
int tc902x_nic_init(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int tc902x_nic_rx(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int tc902x_nic_rxrestore(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int init_rfdlist(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int init_tfdlist(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int tc902x_get_rxbuff(TC902X_DEVICE_TYPE *tc902x_ethernet_device, int rfd);
int tc902x_sti_fiber_detect(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
int tc902x_tmi_fiber_detect(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                            int phyaddr);
int tc902x_config_autoneg(TC902X_DEVICE_TYPE *tc902x_ethernet_device);
unsigned ether_crc_le(int length, unsigned char *data);
int tc902x_nic_do_ioctl(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                        struct ifreq *req, int cmd);
int tc902x_nic_change_mtu(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                          int new_mtu);

#ifdef TC902X_LINUX2_2
int tc902x_pcibussearch_linux2_2(void);
#endif

#ifdef TC902X_LINUX2_4
int tc902x_pciprobe_linux2_4(struct pci_dev*, const struct pci_device_id*);
void tc902x_pciremove_linux2_4(struct pci_dev*);
#endif

/* End function prototypes. */

#ifdef TC902X_DEBUG
void tc902x_dump_rfdlist(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	u32				baseaddr;
	int				i;
	u32				offset;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_dump_rfdlist\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	printk(KERN_INFO "CurrentRFD         = %2.2x\n", sp->CurrentRFD);
	printk(KERN_INFO "LastRestoredRxBuff = %2.2x\n",
	       sp->LastRestoredRxBuff);
#ifdef TC902X_LINUX2_2
	printk(KERN_INFO "RFDList start address = %16.16lx\n",
	       (unsigned long int)(TC902X_HOST2BUS_MAP(sp->RFDList)));
#endif
#ifdef TC902X_LINUX2_4
	printk(KERN_INFO "RFDList start address = %16.16lx\n",
	       (unsigned long int)(sp->RFDListDMAhandle));
#endif
	printk(KERN_INFO "RFDListPtr register   = %8.8x%8.8x\n",
	       TC902X_READ_RFDLISTPTR1(baseaddr),
	       TC902X_READ_RFDLISTPTR0(baseaddr));

	for(i=0;i<TC902X_RFDLIST_LENGTH;i++)
	{
		offset = (u32)(&sp->RFDList[i].RFDNextPtr)
		         - (u32)(sp->RFDList);
		printk(KERN_INFO "%2.2x %4.4x RFDNextPtr = %16.16lx\n", i, 
		       offset, (unsigned long int) sp->RFDList[i].RFDNextPtr);
		offset = (u32)(&sp->RFDList[i].RFS)
		         - (u32)(sp->RFDList);
		printk(KERN_INFO "%2.2x %4.4x RFS        = %16.16lx\n", i, 
		       offset, (unsigned long int)sp->RFDList[i].RFS); 
		offset = (u32)(&sp->RFDList[i].FragInfo)
		         - (u32)(sp->RFDList);
		printk(KERN_INFO "%2.2x %4.4x FragInfo   = %16.16lx\n", i, 
		       offset, (unsigned long int)sp->RFDList[i].FragInfo);
	}

	return;
}

void tc902x_dump_tfdlist(TC902X_DEVICE_TYPE	*tc902x_ethernet_device)
{
	u32				baseaddr;
	int				i;
	u32				offset;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_dump_tfdlist\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	printk(KERN_INFO "CurrentTFD         = %2.2x\n", sp->CurrentTFD);
	printk(KERN_INFO "LastFreedTxBuff = %2.2x\n",
	       sp->LastFreedTxBuff);
#ifdef TC902X_LINUX2_2
	printk(KERN_INFO "TFDList start address = %16.16lx\n",
	       (unsigned long int)(TC902X_HOST2BUS_MAP(sp->TFDList)));
#endif
#ifdef TC902X_LINUX2_4
	printk(KERN_INFO "TFDList start address = %16.16lx\n",
	       (unsigned long int)(sp->TFDListDMAhandle));
#endif
	printk(KERN_INFO "TFDListPtr register   = %8.8x%8.8x\n",
	       TC902X_READ_TFDLISTPTR1(baseaddr),
	       TC902X_READ_TFDLISTPTR0(baseaddr));

	for(i=0;i<TC902X_TFDLIST_LENGTH;i++)
	{
		offset = (u32)(&sp->TFDList[i].TFDNextPtr)
		         - (u32)(sp->TFDList);
		printk(KERN_INFO "%2.2x %4.4x TFDNextPtr = %16.16lx\n", i, 
		       offset, (unsigned long int)sp->TFDList[i].TFDNextPtr);

		offset = (u32)(&sp->TFDList[i].TFC)
		         - (u32)(sp->TFDList);
		printk(KERN_INFO "%2.2x %4.4x TFC        = %16.16lx\n", i, 
		       offset, (unsigned long int)sp->TFDList[i].TFC); 
		offset = (u32)(&sp->TFDList[i].FragInfo)
		         - (u32)(sp->TFDList);
		printk(KERN_INFO "%2.2x %4.4x FragInfo   = %16.16lx\n", i, 
		       offset, (unsigned long int)sp->TFDList[i].FragInfo);
	}

	return;
}
#else /* Not in debug mode. */
#endif

u16	read_phy_register(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
	                  int phy_address, int phy_register)
{
	/* Read a register from the Physical Layer device located
	 * on the TC902X NIC, using the TC902X PHYCTRL register.
	 */

	u32	baseaddr;
	int	i;
	int	j;
	int	fieldlen[8];
	u32	field[8];
	u8	databit;
        u8      phyctrlpolarity;

	TC902X_DEBUG_MSG("read_phy_register\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	/* The GMII mangement frame structure for a read is as follows:
	 *
	 * |Preamble|st|op|phyad|regad|ta|      data      |idle|
	 * |< 32 1s>|01|10|AAAAA|RRRRR|z0|DDDDDDDDDDDDDDDD|z   |
	 * 
	 * <32 1s> = 32 consecutive logic 1 values
	 * A = bit of Physical Layer device address (MSB first)
	 * R = bit of register address (MSB first)
	 * z = High impedance state
	 * D = bit of read data (MSB first)
	 *
	 * Transmission order is 'Preamble' field first, bits transmitted
	 * left to right (first to last).
	 */

	field[0]    = GMII_PREAMBLE;
	fieldlen[0] = 32;		/* Preamble */
	field[1]    = GMII_ST;
	fieldlen[1] = 2;		/* ST */
	field[2]    = GMII_READ;
	fieldlen[2] = 2;		/* OP */
	field[3]    = phy_address;
	fieldlen[3] = 5;		/* PHYAD */
	field[4]    = phy_register;
	fieldlen[4] = 5;		/* REGAD */
	field[5]    = 0x0000;
	fieldlen[5] = 2;		/* TA */
	field[6]    = 0x0000;
	fieldlen[6] = 16;		/* DATA */
	field[7]    = 0x0000;
	fieldlen[7] = 1;		/* IDLE */

        /* Store the polarity values of PHYCTRL. */
        phyctrlpolarity = TC902X_READ_PHYCTRL(baseaddr) &
                          (TC902X_PC_DUPLEX_POLARITY |
	                   TC902X_PC_LINK_POLARITY);
 
	/* Create the Preamble, ST, OP, PHYAD, and REGAD field. */
	for(j=0; j<5; j++)
	for(i=0; i<fieldlen[j]; i++)
	{
		/* For each variable length field, the MSB must be
		 * transmitted first. Rotate through the field bits,
		 * starting with the MSB, and move each bit into the
		 * the 1st (2^1) bit position (this is the bit position
		 * corresponding to the MgmtData bit of the PhyCtrl
		 * register for the TC902X).
		 * 
		 * Example: ST = 01;
		 *
		 *          First write a '0' to bit 1 of the PhyCtrl
		 *          register, then write a '1' to bit 1 of the
		 *          PhyCtrl register.
		 *
		 * To do this, right shift the MSB of ST by the value:
		 * [field length - 1 - #ST bits already written]
		 * then left shift this result by 1.
		 */
		databit = (field[j] >> (fieldlen[j] - 1 - i)) << 1;

		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_LO |
		                     (TC902X_PC_MGMTDATA & databit) |
		                     TC902X_PC_MGMTDIR | phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);
	
		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_HI |
		                     (TC902X_PC_MGMTDATA & databit) |
		                     TC902X_PC_MGMTDIR | phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);
	}

	/* For a read cycle, the bits for the next two fields (TA and
	 * DATA) are driven by the PHY (the TC902X reads these bits).
	 */
	for(j=5; j<8; j++)
	for(i=0; i<fieldlen[j]; i++)
	{
		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_LO |
		                     phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);

		/* Prior to the rising edge of MGMTCLK, read the
		 * value of MGMTDATA which corresponds to a single
		 * bit of the desired value (MSB first). The
		 * value is built by ORing the 16 bits, read one
		 * at a time and shift to the correct position.
		 * The bit shifting is performed as follows:
		 *
		 * Shift the read bit to the right by 1 (to account for
		 * the location, bit position 1, or 2^1 of MGMTDATA
		 * within the PHYCTRL register). Then shift the read
		 * bit (now at location 2^0) to the left by the number
		 * of bits corresponding to it's position in the final
		 * value (i.e. the first read bit is the MSB or the bit
		 * at the 2^16 position, so shift this bit 15 positions
		 * to the left).
		 */ 
		field[j] |= ((TC902X_READ_PHYCTRL(baseaddr) &
		             TC902X_PC_MGMTDATA) >> 1)
		            << (fieldlen[j] - 1 - i);

		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_HI |
		                     phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);

	}

	/* Return the value of the DATA field. */
	return field[6];
}

void	write_phy_register(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
	                  int phy_address, int phy_register, u16 writeval)
{
	/* Write to a register from the Physical Layer device located
	 * on the TC902X NIC, using the TC902X PHYCTRL register.
	 */

	u32	baseaddr;
	int	i;
	int	j;
	int	fieldlen[8];
	u32	field[8];
	u8	databit;
	u8	phyctrlpolarity;

	TC902X_DEBUG_MSG("write_phy_register\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	/* The GMII mangement frame structure for a read is as follows:
	 *
	 * |Preamble|st|op|phyad|regad|ta|      data      |idle|
	 * |< 32 1s>|01|10|AAAAA|RRRRR|z0|DDDDDDDDDDDDDDDD|z   |
	 * 
	 * <32 1s> = 32 consecutive logic 1 values
	 * A = bit of Physical Layer device address (MSB first)
	 * R = bit of register address (MSB first)
	 * z = High impedance state
	 * D = bit of write data (MSB first)
	 *
	 * Transmission order is 'Preamble' field first, bits transmitted
	 * left to right (first to last).
	 */

	field[0]    = GMII_PREAMBLE;
	fieldlen[0] = 32;		/* Preamble */
	field[1]    = GMII_ST;
	fieldlen[1] = 2;		/* ST */
	field[2]    = GMII_WRITE;
	fieldlen[2] = 2;		/* OP */
	field[3]    = phy_address;
	fieldlen[3] = 5;		/* PHYAD */
	field[4]    = phy_register;
	fieldlen[4] = 5;		/* REGAD */
	field[5]    = 0x0002;
	fieldlen[5] = 2;		/* TA */
	field[6]    = writeval;
	fieldlen[6] = 16;		/* DATA */
	field[7]    = 0x0000;
	fieldlen[7] = 1;		/* IDLE */

        /* Store the polarity values of PHYCTRL. */
        phyctrlpolarity = TC902X_READ_PHYCTRL(baseaddr) &
                          (TC902X_PC_DUPLEX_POLARITY |
	                   TC902X_PC_LINK_POLARITY);

	/* Create the Preamble, ST, OP, PHYAD, and REGAD field. */
	for(j=0; j<7; j++)
	for(i=0; i<fieldlen[j]; i++)
	{
		/* For each variable length field, the MSB must be
		 * transmitted first. Rotate through the field bits,
		 * starting with the MSB, and move each bit into the
		 * the 1st (2^1) bit position (this is the bit position
		 * corresponding to the MgmtData bit of the PhyCtrl
		 * register for the TC902X).
		 * 
		 * Example: ST = 01;
		 *
		 *          First write a '0' to bit 1 of the PhyCtrl
		 *          register, then write a '1' to bit 1 of the
		 *          PhyCtrl register.
		 *
		 * To do this, right shift the MSB of ST by the value:
		 * [field length - 1 - #ST bits already written]
		 * then left shift this result by 1.
		 */
		databit = (field[j] >> (fieldlen[j] - 1 - i)) << 1;

		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_LO |
		                     (TC902X_PC_MGMTDATA & databit) |
		                     TC902X_PC_MGMTDIR | phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);
	
		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_HI |
		                     (TC902X_PC_MGMTDATA & databit) |
		                     TC902X_PC_MGMTDIR | phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);
	}

	/* The last cycle is a tri-state, so read from the PHY.
	 */
	for(j=7; j<8; j++)
	for(i=0; i<fieldlen[j]; i++)
	{
		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_LO |
		                     phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);

		field[j] |= ((TC902X_READ_PHYCTRL(baseaddr) &
		             TC902X_PC_MGMTDATA) >> 1)
		            << (fieldlen[j] - 1 - i);

		TC902X_WRITE_PHYCTRL(baseaddr, TC902X_PC_MGMTCLK_HI |
		                     phyctrlpolarity);

		mdelay(TC902X_PC_PHYCTRLWAIT);

	}

	return;
}

int	tc902x_reset(u32 baseaddr, u32 resetflags)
{
	/* Assert functional resets via the TC902X AsicCtrl
	 * register as specified by the 'resetflags' input
	 * parameter.
	 */

	int	timeout_count;

	TC902X_DEBUG_MSG("_reset\n");

	TC902X_WRITE_ASICCTRL(baseaddr, TC902X_READ_ASICCTRL(baseaddr) |
	               resetflags);

	/* Wait until TC902X reset is complete. */
	timeout_count = 0;

	/* Delay added to account for problem with 10Mbps reset. */
	mdelay(TC902X_AC_RESETWAIT);

	while (TC902X_AC_RESET_BUSY & TC902X_READ_ASICCTRL(baseaddr))
	{
		mdelay(TC902X_AC_RESETWAIT);
		timeout_count++;
		if (timeout_count > TC902X_AC_RESET_TIMEOUT)
			return -ETIME;
	}

	return 0;
}

int	tc902x_sti_fiber_detect(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Determine if NIC is fiber based by reading the PhyMedia
	 * bit in the AsicCtrl register.
	 */

	u32				asicctrl;
	u32				baseaddr;

	TC902X_DEBUG_MSG("_sti_fiber_detect\n");

	baseaddr = tc902x_ethernet_device->base_addr;
	asicctrl = TC902X_READ_ASICCTRL(baseaddr);

        if (asicctrl & TC902X_AC_PHY_MEDIA)
	{
		/* Fiber NIC. */
		return 1;
	} else
	{
		/* Not a fiber NIC. */
		return 0;
	}
}

int	tc902x_tmi_fiber_detect(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
	                        int phyaddr)
{
	/* Determine if NIC is fiber based by reading the ID register 
	 * of the PHY and the GMII address.
	 */

	u16				phyid;
	u32				baseaddr;

	TC902X_DEBUG_MSG("_tmi_fiber_detect\n");

	baseaddr = tc902x_ethernet_device->base_addr;
	phyid = read_phy_register(tc902x_ethernet_device,
	                          phyaddr, GMII_PHY_ID_1);

	TC902X_DEBUG_MSG("PHY ID = %x\n", phyid);

	/* We conclude the mode is fiber if the GMII address
	 * is 0x1 and the PHY ID is 0x0000.
	 */
        if ((phyaddr == 0x1) && (phyid == 0x0000))
	{
		/* Fiber NIC. */
		return 1;
	} else
	{
		/* Not a fiber NIC. */
		return 0;
	}
}

int	tc902x_find_phyaddr(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Find the GMII PHY address. */

	int	i;
	int	phyaddr;
	u32	status;

        for(i=0;i<32;i++)
        {
                /* Search for the correct PHY address among 32 possible. */
                phyaddr = (TC902X_NIC_PHY_ADDRESS + i) % 32;
 
                status = read_phy_register(tc902x_ethernet_device,
                                           phyaddr, GMII_PHY_STATUS);
 
                if (status != 0xFFFF)
                        return phyaddr;
        }

	return -1;
}

int	tc902x_config_autoneg(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Configure TC902X based on result of IEEE 802.3 PHY
	 * auto-negotiation.
	 */

        int                             phyaddr = 0;
	u8				phyctrl;
	u32				asicctrl;
	u32				baseaddr;
        u16                             status = 0;
	u16				advertisement;
	u16				linkpartner_ability;
	u16				gigadvertisement;
	u16				giglinkpartner_ability;
        u16                             techabilities;
        int                             fiber;
        int                             gig;
        int                             fullduplex;
        int                             txflowcontrol;
        int                             rxflowcontrol;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_config_autoneg\n");

	baseaddr = tc902x_ethernet_device->base_addr;
	asicctrl = TC902X_READ_ASICCTRL(baseaddr);
        phyctrl = TC902X_READ_PHYCTRL(baseaddr);

	/* Set flags for use in resolving auto-negotation, assuming
	 * non-1000Mbps, half duplex, no flow control.
	 */
	fiber = 0;
	fullduplex = 0;
	txflowcontrol = 0;
	rxflowcontrol = 0;
	gig = 0;

	/* To accomodate a problem in 10Mbps operation,
	 * set a global flag if PHY running in 10Mbps mode.
	 */
 	sp->tenmbpsmode = 0;

	printk("Link speed = ");

	/* Determine actual speed of operation. */
	switch (phyctrl & TC902X_PC_LINK_SPEED)
	{
		case TC902X_PC_LINK_SPEED_10MBPS :
			printk("10Mbps.\n");
                	printk(KERN_INFO "%s: 10Mbps operational mode enabled.\n",tc902x_ethernet_device->name);
	 		sp->tenmbpsmode = 1;
			break;
		case TC902X_PC_LINK_SPEED_100MBPS :
			printk("100Mbps.\n");
			break;
		case TC902X_PC_LINK_SPEED_1000MBPS :
			printk("1000Mbps.\n");
			gig = 1;
			break;
		default : printk("undefined!\n");
	}

#ifndef TC902X_TMI_FIBER_DETECT
	fiber = tc902x_sti_fiber_detect(tc902x_ethernet_device);

        /* Determine if auto-negotiation resolution is necessary.
	 * First check for fiber based media 10/100 media.
	 */
        if ((fiber == 1) && (asicctrl &
	    (TC902X_AC_PHY_SPEED10 | TC902X_AC_PHY_SPEED100)))
        {
                printk(KERN_INFO "%s: Fiber based PHY, setting full duplex, no flow control.\n", tc902x_ethernet_device->name);
                return -EILSEQ;
                TC902X_WRITE_MACCTRL(baseaddr, (TC902X_READ_MACCTRL(baseaddr) |
                                                TC902X_MC_DUPLEX_SELECT_FD) &
                                     ~TC902X_MC_TX_FLOW_CONTROL_ENABLE &
                                     ~TC902X_MC_RX_FLOW_CONTROL_ENABLE);
 
                return 0;
        }
#endif

        /* Determine if PHY is auto-negotiation capable. */
	phyaddr = tc902x_find_phyaddr(tc902x_ethernet_device);

        if (phyaddr == -1)
        {
                printk(KERN_INFO "%s: Error on read to GMII/MII Status register.\n",tc902x_ethernet_device->name);
                return -EILSEQ;
        }
 
	TC902X_DEBUG_MSG("GMII/MII PHY address = %x\n", phyaddr);

        status = read_phy_register(tc902x_ethernet_device,
                                   phyaddr, GMII_PHY_STATUS);

        if ((status & GMII_PHY_STATUS_AUTONEG_ABILITY) == 0)
        {
                printk(KERN_INFO "%s: Error PHY unable to perform auto-negotiation.\n",
			tc902x_ethernet_device->name);
                return -EILSEQ;
        }

	advertisement = read_phy_register(tc902x_ethernet_device, phyaddr,
	                                  GMII_PHY_AUTONEGADVERTISEMENT);
	linkpartner_ability = read_phy_register(tc902x_ethernet_device, phyaddr,
	                                        GMII_PHY_AUTONEGLINKPARTABILITY);

	if ((advertisement == 0xFFFF) || (linkpartner_ability == 0xFFFF))
	{
		printk(KERN_INFO "%s: Error on read to GMII/MII registers 4 and/or 5.\n", tc902x_ethernet_device->name);
		return -EILSEQ;
	}

#ifdef TC902X_TMI_FIBER_DETECT
	fiber = tc902x_tmi_fiber_detect(tc902x_ethernet_device, phyaddr);
#endif

        /* Resolve full/half duplex if 1000BASE-X. */
	if ((gig == 1) && (fiber == 1))
        {
		/* Compare the full duplex bits in the GMII registers
		 * for the local device, and the link partner. If these
		 * bits are logic 1 in both registers, configure the
		 * TC902X for full duplex operation.
		 */
		if ((advertisement & GMII_PHY_ADV_FULL_DUPLEX) ==
		    (linkpartner_ability & GMII_PHY_ADV_FULL_DUPLEX))
		{
			fullduplex = 1;

			/* In 1000BASE-X using TC902X's internal PCS
			 * layer, so write to the GMII duplex bit.	
			 */
			write_phy_register(tc902x_ethernet_device,
			                   phyaddr,
			                   GMII_PHY_CONTROL,
			                   read_phy_register(tc902x_ethernet_device,
			                   phyaddr,
			                   GMII_PHY_CONTROL) |
			                   GMII_PHY_CONTROL_FULL_DUPLEX);
	
		} else
		{
			fullduplex = 0;

			/* In 1000BASE-X using TC902X's internal PCS
			 * layer, so write to the GMII duplex bit.	
			 */
			write_phy_register(tc902x_ethernet_device,
			                   phyaddr,
			                   GMII_PHY_CONTROL,
			                   read_phy_register(tc902x_ethernet_device,
			                   phyaddr,
			                   GMII_PHY_CONTROL) &
			                   ~GMII_PHY_CONTROL_FULL_DUPLEX);
		}
	}

        /* Resolve full/half duplex if 1000BASE-T. */
	if ((gig == 1) && (fiber == 0))
        {
		/* Read the 1000BASE-T "Control" and "Status"
		 * registers which represent the advertised and
		 * link partner abilities exchanged via next page
		 * transfers.
		 */
		gigadvertisement = read_phy_register(tc902x_ethernet_device,
		                                     phyaddr,
	                                         GMII_PHY_1000BASETCONTROL);
		giglinkpartner_ability = read_phy_register(tc902x_ethernet_device,
		                                           phyaddr,
	                                          GMII_PHY_1000BASETSTATUS);

		/* Compare the full duplex bits in the 1000BASE-T GMII
		 * registers for the local device, and the link partner.
		 * If these bits are logic 1 in both registers, configure
		 * the TC902X for full duplex operation.
		 */
		if ((gigadvertisement & GMII_PHY_1000BASETCONTROL_FULL_DUPLEX) &&
		    (giglinkpartner_ability & GMII_PHY_1000BASETSTATUS_FULL_DUPLEX))
		{
			fullduplex = 1;
		} else
		{
			fullduplex = 0;
		}
	}
	
        /* Resolve full/half duplex for 10/100BASE-T. */
	if (gig == 0)
	{
	        /* Autonegotiation Priority Resolution algorithm, as defined in
	       	 * IEEE 802.3 Annex 28B.
	       	 */
	        if (((advertisement & MII_PHY_SELECTORFIELD) ==
	             MII_PHY_SELECTOR_IEEE8023) &&
	            ((linkpartner_ability & MII_PHY_SELECTORFIELD) ==
	             MII_PHY_SELECTOR_IEEE8023))
	        {
	                techabilities = (advertisement & linkpartner_ability &
	                                 MII_PHY_TECHABILITYFIELD);
 
	                fullduplex = 0;
 
	                /* 10BASE-TX half duplex is lowest priority. */
	                if (techabilities & MII_PHY_TECHABILITY_10BT)
	                {
	                        fullduplex = 0;
	                }
 
	                if (techabilities & MII_PHY_TECHABILITY_10BTFD)
	                {
	                        fullduplex = 1;
	                }
 
	                if (techabilities & MII_PHY_TECHABILITY_100BTX)
	                {
	                        fullduplex = 0;
	                }
 
	                if (techabilities & MII_PHY_TECHABILITY_100BT4)
	                {
	                        fullduplex = 0;
	                }
	 
	                /* 100BASE-TX half duplex is highest priority. */ //Sorbica full duplex ?
	                if (techabilities & MII_PHY_TECHABILITY_100BTXFD)
	                {
	                        fullduplex = 1;
	                }
	
	                if (fullduplex == 1)
	                {
	                        /* If in full duplex mode, determine if PAUSE
	                         * functionality is supported by the local
				 * device, and the link partner.
	                         */
	                        if (techabilities & MII_PHY_TECHABILITY_PAUSE)
	                        {
					txflowcontrol = 1;
					rxflowcontrol = 1;
	                        }
	                        else
	                        {
					txflowcontrol = 0;
					rxflowcontrol = 0;
	                        }
	                }
	        }
	}

	/* If in 1000Mbps, fiber, and full duplex mode, resolve
	 * 1000BASE-X PAUSE capabilities. */
	if ((fullduplex == 1) && (fiber == 1) && (gig == 1))
	{
		/* In full duplex mode, resolve PAUSE
		 * functionality.
		 */
		switch(((advertisement & GMII_PHY_ADV_PAUSE) >> 5) |
		       ((linkpartner_ability & GMII_PHY_ADV_PAUSE) >> 7))
		{
			case 0x7 :
			txflowcontrol = 1;
			rxflowcontrol = 0;
			break;

			case 0xA : case 0xB: case 0xE: case 0xF:
			txflowcontrol = 1;
			rxflowcontrol = 1;
			break;

			case 0xD :
			txflowcontrol = 0;
			rxflowcontrol = 1;
			break;

			default :
			txflowcontrol = 0;
			rxflowcontrol = 0;
		}
	}

	/* If in 1000Mbps, non-fiber, full duplex mode, resolve
	 * 1000BASE-T PAUSE capabilities. */
	if ((fullduplex == 1) && (fiber == 0) && (gig == 1))
	{
		/* Make sure the PHY is advertising we are PAUSE
		 * capable.
		 */
		if (!(advertisement & (MII_PHY_TECHABILITY_PAUSE |
		                       MII_PHY_TECHABILITY_ASM_DIR)))
		{
			/* PAUSE is not being advertised. Advertise
			 * PAUSE and restart auto-negotiation.
			 */
			write_phy_register(tc902x_ethernet_device,
		                           phyaddr,
			                   MII_PHY_AUTONEGADVERTISEMENT,
			                   (advertisement |
                                            MII_PHY_TECHABILITY_PAUSE |
			                    MII_PHY_TECHABILITY_ASM_DIR));
			write_phy_register(tc902x_ethernet_device,
		                           phyaddr,
			                   MII_PHY_CONTROL,
			                   MII_PHY_CONTROL_RESTARTAN);

			return -EAGAIN;
		}

		/* In full duplex mode, resolve PAUSE
		 * functionality.
		 */
		switch(((advertisement &
		         MII_PHY_TECHABILITY_PAUSE_FIELDS) >> 0x8) |
		       ((linkpartner_ability &
		         MII_PHY_TECHABILITY_PAUSE_FIELDS) >> 0xA))
		{
			case 0x7 :
			txflowcontrol = 1;
			rxflowcontrol = 0;
			break;

			case 0xA : case 0xB: case 0xE: case 0xF:
			txflowcontrol = 1;
			rxflowcontrol = 1;
			break;

			case 0xD :
			txflowcontrol = 0;
			rxflowcontrol = 1;
			break;

			default :
			txflowcontrol = 0;
			rxflowcontrol = 0;
		}
	}

	/* If in 10/100Mbps, non-fiber, full duplex mode, assure
	 * 10/100BASE-T PAUSE capabilities are advertised. */
	if ((fullduplex == 1) && (fiber == 0) && (gig == 0))
	{
		/* Make sure the PHY is advertising we are PAUSE
		 * capable.
		 */
		if (!(advertisement & (MII_PHY_TECHABILITY_PAUSE)))
		{
			/* PAUSE is not being advertised. Advertise
			 * PAUSE and restart auto-negotiation.
			 */
			write_phy_register(tc902x_ethernet_device,
		                           phyaddr,
			                   MII_PHY_AUTONEGADVERTISEMENT,
			                   (advertisement |
                                            MII_PHY_TECHABILITY_PAUSE));
			write_phy_register(tc902x_ethernet_device,
		                           phyaddr,
			                   MII_PHY_CONTROL,
			                   MII_PHY_CONTROL_RESTARTAN);

			return -EAGAIN;
		}

	}

        if (fiber == 1)
	{
        	printk(KERN_INFO "%s: Fiber based PHY, ",
		       tc902x_ethernet_device->name);
	} else
	{
        	printk(KERN_INFO "%s: Copper based PHY, ",
		       tc902x_ethernet_device->name);
	}

	/* Configure full duplex, and flow control. */
	if (fullduplex == 1)
	{
		/* Configure TC902X for full duplex operation. */
        	printk("setting full duplex, ");

		TC902X_WRITE_MACCTRL(baseaddr, TC902X_READ_MACCTRL(baseaddr) |
	                             TC902X_MC_DUPLEX_SELECT_FD);

		if (txflowcontrol == 1)
		{
        		printk("TX flow control");
			TC902X_WRITE_MACCTRL(baseaddr,
			                     TC902X_READ_MACCTRL(baseaddr) |
			                     TC902X_MC_TX_FLOW_CONTROL_ENABLE);
		} else
		{
        		printk("no TX flow control");
			TC902X_WRITE_MACCTRL(baseaddr,
			                     TC902X_READ_MACCTRL(baseaddr) &
			                     ~TC902X_MC_TX_FLOW_CONTROL_ENABLE);
		}

		if (rxflowcontrol == 1)
		{
        		printk(", RX flow control.");
			TC902X_WRITE_MACCTRL(baseaddr,
			                     TC902X_READ_MACCTRL(baseaddr) |
			                     TC902X_MC_RX_FLOW_CONTROL_ENABLE);
		} else
		{
        		printk(", no RX flow control.");
			TC902X_WRITE_MACCTRL(baseaddr,
			                     TC902X_READ_MACCTRL(baseaddr) &
			                     ~TC902X_MC_RX_FLOW_CONTROL_ENABLE);
		}

		printk("\n");
	} else
	{
		/* Configure TC902X for half duplex operation. */
	        printk("setting half duplex, no TX flow control, no RX flow control.\n");

		TC902X_WRITE_MACCTRL(baseaddr, TC902X_READ_MACCTRL(baseaddr) &
		                     ~TC902X_MC_DUPLEX_SELECT_FD &
		                     ~TC902X_MC_TX_FLOW_CONTROL_ENABLE &
		                     ~TC902X_MC_RX_FLOW_CONTROL_ENABLE);
	}
	

	TC902X_DEBUG_MSG("G/MII reg 4 (advertisement) = %4.4x\n", advertisement);
	TC902X_DEBUG_MSG("G/MII reg 5 (link partner)  = %4.4x\n", linkpartner_ability);
	TC902X_DEBUG_MSG("G/MII reg 9 (1000BASE-T control) = %4.4x\n", advertisement);
	TC902X_DEBUG_MSG("G/MII reg 10 (1000BASE-T status) = %4.4x\n", linkpartner_ability);

	TC902X_DEBUG_MSG("Auto-neg complete, MACCTRL = %8.8x\n",
	          TC902X_READ_MACCTRL(baseaddr));

	return 0;
}

int	tc902x_io_config(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Initialize the TC902X I/O registers. */

	u32		baseaddr;
	u32		origmacctrl;
	u32		restoremacctrl;

	TC902X_DEBUG_MSG("_io_config\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	/* Save the original value of MACCTRL. */
	origmacctrl = TC902X_READ_MACCTRL(baseaddr);

	/* Establish a vlaue to restore MACCTRL when done. */
	restoremacctrl = origmacctrl;

	/* Enable statistics gathering. */
	restoremacctrl |= TC902X_MC_STATISTICS_ENABLE;

        /* Based on compilation option, determine if FCS is to be
         * stripped on receive frames by TC902X.
         */
        if (!(TC902X_STRIP_FCS_ON_RX))
        {
		restoremacctrl |= TC902X_MC_RCV_FCS;
        }

	/* Determine if transmitter and/or receiver are
	 * enabled so we may restore MACCTRL correctly.
	 */
	if (origmacctrl & TC902X_MC_TX_ENABLED)
	{
		restoremacctrl |= TC902X_MC_TX_ENABLE;
	}

	if (origmacctrl & TC902X_MC_RX_ENABLED)
	{
		restoremacctrl |= TC902X_MC_RX_ENABLE;
	}

	/* Transmitter and receiver must be disabled before setting
	 * IFSSelect. 
	 */
	TC902X_WRITE_MACCTRL(baseaddr, origmacctrl & (TC902X_MC_RX_DISABLE |
	                     TC902X_MC_TX_DISABLE));

	/* Now that transmitter and receiver are disabled, write
	 * to IFSSelect.
	 */
	TC902X_WRITE_MACCTRL(baseaddr, origmacctrl & TC902X_MC_IFS_96BIT);

	/* Set RECEIVEMODE register. */
	tc902x_nic_set_multicast_list(tc902x_ethernet_device);

	TC902X_WRITE_MAXFRAMESIZE(baseaddr, TC902X_MAX_RXFRAME_SIZE);
	TC902X_WRITE_RXEARLYTHRESH(baseaddr, TC902X_RXEARLYTHRESH_VALUE);
	TC902X_WRITE_TXSTARTTHRESH(baseaddr, TC902X_TXSTARTTHRESH_VALUE);
	TC902X_WRITE_RXDMAINTCTRL(baseaddr, (TC902X_RI_RSVD_MASK &
	                              ((TC902X_RI_RXFRAME_COUNT &
	                                TC902X_RXFRAME_COUNT) |
	                               (TC902X_RI_PRIORITY_THRESH &
	                                (TC902X_PRIORITY_THRESH << 12)) |
	                               (TC902X_RI_RXDMAWAIT_TIME &
	                                (TC902X_RXDMAWAIT_TIME << 16)))));
	TC902X_WRITE_RXDMAPOLLPERIOD(baseaddr, TC902X_RXDMAPOLLPERIOD_VALUE);
	TC902X_WRITE_RXDMAURGENTTHRESH(baseaddr,
	                               TC902X_RXDMAURGENTTHRESH_VALUE);
	TC902X_WRITE_RXDMABURSTTHRESH(baseaddr,
	                              TC902X_RXDMABURSTTHRESH_VALUE);
	TC902X_WRITE_TXDMAPOLLPERIOD(baseaddr,
	                             TC902X_TXDMAPOLLPERIOD_VALUE);
	TC902X_WRITE_TXDMAURGENTTHRESH(baseaddr,
	                               TC902X_TXDMAURGENTTHRESH_VALUE);
	TC902X_WRITE_TXDMABURSTTHRESH(baseaddr,
	                              TC902X_TXDMABURSTTHRESH_VALUE);
	TC902X_WRITE_INTENABLE(baseaddr, TC902X_IE_HOST_ERROR |
	                          TC902X_IE_TX_DMA_COMPLETE |
	                          TC902X_IE_TX_COMPLETE |
	                          TC902X_IE_INT_REQUESTED |
	                          TC902X_IE_UPDATE_STATS |
	                          TC902X_IE_LINK_EVENT |
	                          TC902X_IE_RX_DMA_COMPLETE |
	                          TC902X_IE_RFD_LIST_END |
	                          TC902X_IE_RX_DMA_PRIORITY);

	TC902X_WRITE_FLOWONTHRESH(baseaddr, TC902X_FLOWONTHRESH_VALUE);
	TC902X_WRITE_FLOWOFFTHRESH(baseaddr, TC902X_FLOWOFFTHRESH_VALUE);

	/* TC902X multi-frag frame bug workaround.
	 * Per silicon revision B3 eratta.
	 */
	TC902X_WRITE_DEBUGCTRL(baseaddr,
	                       TC902X_READ_DEBUGCTRL(baseaddr) | 0x0200);

	/* TC902X TX poll now bug workaround.
	 * Per silicon revision B3 eratta.
	 */
	TC902X_WRITE_DEBUGCTRL(baseaddr,
	                       TC902X_READ_DEBUGCTRL(baseaddr) | 0x0010);

	/* TC902X RX poll now bug workaround.
	 * Per silicon revision B3 eratta.
	 */
	TC902X_WRITE_DEBUGCTRL(baseaddr,
	                       TC902X_READ_DEBUGCTRL(baseaddr) | 0x0020);

	/* Now restore MACCTRL to original setting. */
	TC902X_WRITE_MACCTRL(baseaddr, restoremacctrl);

	/* Disable unused RMON statistics. */
	TC902X_WRITE_RMONSTATISTICSMASK(baseaddr, TC902X_RZ_ALL);

	/* Disable unused MIB statistics. */
	TC902X_WRITE_STATISTICSMASK(baseaddr,
	                     TC902X_SM_MACCONTROLFRAMESXMTD |
	                     TC902X_SM_BCSTOCTETXMTOK_BCSTFRAMESXMTDOK |
	                     TC902X_SM_MCSTOCTETXMTOK_MCSTFRAMESXMTDOK |
	                     TC902X_SM_MACCONTROLFRAMESRCVD |
	                     TC902X_SM_BCSTOCTETRCVDOK_BCSTFRAMESRCVDOK |
	                     TC902X_SM_TXJUMBOFRAMES |
	                     TC902X_SM_UDPCHECKSUMERRORS |
	                     TC902X_SM_IPCHECKSUMERRORS |
	                     TC902X_SM_TCPCHECKSUMERRORS |
	                     TC902X_SM_RXJUMBOFRAMES);

	return 0;
}

void	tc902x_interrupt_handler(int tc902x_irq, void *device_instance,
	                        struct pt_regs *regs) 
{
	int				error;
	u32				baseaddr;
	u16				intstatusackword;
        TC902X_DEVICE_TYPE		*tc902x_ethernet_device =
		                        (TC902X_DEVICE_TYPE *)device_instance;
#ifdef TC902X_DEBUG
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;
#endif

	TC902X_DEBUG_MSG("_interrupt_handler\n");

	baseaddr = tc902x_ethernet_device->base_addr;

#ifdef TC902X_LINUX2_2
        /*
         * The following code fragment was authored by Donald Becker.
         */
#if defined(__i386__)
                /* A lock to prevent simultaneous entry bug on Intel SMP machines. */
                if (test_and_set_bit(0, (void*)&tc902x_ethernet_device->interrupt)) {
                        printk(KERN_ERR "%s: SMP simultaneous entry of an interrupt handler.\n",
                               tc902x_ethernet_device->name);
                        tc902x_ethernet_device->interrupt = 0;     /* Avoid halting machine. */
                        return;
                }
#else
                if (tc902x_ethernet_device->interrupt) {
                        printk(KERN_ERR "%s: Re-entering the interrupt handler.\n",
                               tc902x_ethernet_device->name);
                        return;
                }
                tc902x_ethernet_device->interrupt = 1;
#endif
        /*
         * End of code fragment authored by Donald Becker.
         */

        /* Setting tc902x_ethernet_device->interrupt = 1
         * indicates to higher network layers an
         * Ethernet NIC interrupt is being serviced.
         */
#endif

	/* Get interrupt source information, and acknowledge
	 * some (i.e. TxDMAComplete, RxDMAComplete, RxEarly,
	 * IntRequested, MacControlFrame, LinkEvent) interrupts
	 * if issued. Also, all TC902X interrupts are disabled by
	 * reading IntStatusAck.
	 */
	intstatusackword = TC902X_READ_INTSTATUSACK(baseaddr);

	TC902X_DEBUG_MSG("IntStatusAck = %4.4x\n", intstatusackword);

	/* If RFDListEnd interrupt, restore all used RFDs. */
	if (intstatusackword & TC902X_IS_RFD_LIST_END)
	{
		TC902X_DEBUG_MSG("RFDListEnd Interrupt.\n");

		/* The RFD list end indicates an RFD was encountered
		 * with a 0 NextPtr, or with an RFDDone bit set to 1
		 * (indicating the RFD is not read for use by the
		 * TC902X.) Try to restore all RFDs.
		 */
		tc902x_nic_rxrestore(tc902x_ethernet_device);

#ifdef TC902X_DEBUG
		/* Increment the RFDlistendCount counter. */
		sp->RFDlistendCount++;
#endif
	}

	/* If RFDListEnd, RxDMAPriority, RxDMAComplete, or
	 * IntRequested interrupt, process received frames. */
	if ((intstatusackword & TC902X_IS_RX_DMA_PRIORITY) ||
	    (intstatusackword & TC902X_IS_RFD_LIST_END) ||
	    (intstatusackword & TC902X_IS_RX_DMA_COMPLETE) ||
	    (intstatusackword & TC902X_IS_INT_REQUESTED))
	{

#ifdef TC902X_DEBUG
	/* Increment the RFD list checked counter if interrupted
	 * only to check the RFD list. */
	if (intstatusackword & (~(TC902X_IS_RX_DMA_PRIORITY |
	    TC902X_IS_RFD_LIST_END | TC902X_IS_RX_DMA_COMPLETE |
	    TC902X_IS_INT_REQUESTED) &
	                          (TC902X_IS_HOST_ERROR |
	                          TC902X_IS_TX_DMA_COMPLETE |
	                          TC902X_IS_TX_COMPLETE |
	                          TC902X_IS_UPDATE_STATS |
	                          TC902X_IS_LINK_EVENT)))

	{
		sp->RFDListCheckedCount++;
	}
#endif

		tc902x_nic_rx(tc902x_ethernet_device);
	}

	/* If TxDMAComplete interrupt, free used TFDs. */
	if (intstatusackword & TC902X_IS_TX_DMA_COMPLETE)
	{
		/* Free used TFDs. */
		tc902x_nic_txfree(tc902x_ethernet_device);
	}

	/* TxComplete interrupts indicate one of numerous actions.
	 * Determine what action to take based on TXSTATUS register.
	 */
	if (intstatusackword & TC902X_IS_TX_COMPLETE)
	{
		tc902x_nic_txcleanup(tc902x_ethernet_device);
	}

	/* If UpdateStats interrupt, update Linux Ethernet statistics */
	if (intstatusackword & TC902X_IS_UPDATE_STATS)
	{
		tc902x_nic_get_stats(tc902x_ethernet_device);
	}

	/* If HostError interrupt, reset TC902X. */
	if (intstatusackword & TC902X_IS_HOST_ERROR)
	{
		TC902X_DDEBUG_MSG("HostError Interrupt\n");

		TC902X_DDEBUG_MSG("DMACtrl = %8.8x\n",
		 TC902X_READ_DMACTRL(baseaddr));

		/* Acknowledge HostError interrupt by resetting
		 * TC902X DMA and HOST.
		 */
		tc902x_reset(baseaddr,
		            TC902X_AC_GLOBAL_RESET |
		            TC902X_AC_HOST |
		            TC902X_AC_DMA);

		error = tc902x_io_config(tc902x_ethernet_device);
		if (error < 0)
		{
                	printk(KERN_INFO "%s: Cannot recover from PCI error.\n",
			       tc902x_ethernet_device->name);
		}

		init_rfdlist(tc902x_ethernet_device);

		init_tfdlist(tc902x_ethernet_device);
	}

	/* If LinkEvent interrupt, resolve autonegotiation. */
	if (intstatusackword & TC902X_IS_LINK_EVENT)
	{
		if (tc902x_config_autoneg(tc902x_ethernet_device) < 0)
			printk(KERN_INFO "%s: Auto-negotiation error.\n",
			       tc902x_ethernet_device->name);

	}

	/* If MACCtrlFrame interrupt, do nothing. */
	if (intstatusackword & TC902X_IS_MAC_CTRL_FRAME)
	{
		TC902X_DEBUG_MSG("MACCtrlFrame interrupt.\n");
	}

	/* If RxComplete interrupt, do nothing. */
	if (intstatusackword & TC902X_IS_RX_COMPLETE)
	{
		TC902X_DEBUG_MSG("RxComplete interrupt.\n");
	}

	/* If RxEarly interrupt, do nothing. */
	if (intstatusackword & TC902X_IS_RX_EARLY)
	{
		TC902X_DEBUG_MSG("RxEarly interrupt.\n");
	}

	/* Re-enable TC902X interrupts. */
	TC902X_WRITE_INTENABLE(baseaddr, TC902X_IE_HOST_ERROR |
	                          TC902X_IE_TX_DMA_COMPLETE |
	                          TC902X_IE_TX_COMPLETE |
	                          TC902X_IE_INT_REQUESTED |
	                          TC902X_IE_UPDATE_STATS |
	                          TC902X_IE_LINK_EVENT |
	                          TC902X_IE_RX_DMA_COMPLETE |
	                          TC902X_IE_RFD_LIST_END |
	                          TC902X_IE_RX_DMA_PRIORITY);

        /* Indicate to higher network layers the Ethernet NIC
         * interrupt servicing is complete.
         */

#ifdef TC902X_LINUX2_2
        /*
         * The following code fragment was authored by Donald Becker.
         */
#if defined(__i386__)
                clear_bit(0, (void*)&tc902x_ethernet_device->interrupt);
#else
                tc902x_ethernet_device->interrupt = 0;
#endif
        /*
         * End of code fragment authored by Donald Becker.
         */
#endif

        return;
}

void tc902x_nic_txcleanup(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* For TxComplete interrupts, free all transmit
	 * buffers which have already been transfered via DMA
	 * to the TC902X.
	 */

	int				maxtfdcount;
	u32				baseaddr;
	u32				txstatusdword;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_nic_txcleanup\n");

	baseaddr = tc902x_ethernet_device->base_addr;
	maxtfdcount = TC902X_TFDLIST_LENGTH;

	do
	{
		/* Reading the TXSTATUS register clears the
		 * TX_COMPLETE interrupt.
		 */
		txstatusdword = TC902X_READ_TXSTATUS(baseaddr);

		TC902X_DEBUG_MSG("TxStatus = %8.8x\n",txstatusdword);

		/* Check for Transmit errors. Error bits only valid if
		 * TX_COMPLETE bit in the TXSTATUS register is a 1.
		 */
		if (txstatusdword & TC902X_TS_TX_COMPLETE)
		{

			/* If in 10Mbps mode, indicate transmit is ready. */
			if (sp->tenmbpsmode)
			{
				spin_lock(&sp->lock);
				TC902X_TX_NOTBUSY(tc902x_ethernet_device);
				spin_unlock(&sp->lock);
			}

			/* Transmit error, increment stat counters. */
			if (txstatusdword & TC902X_TS_TX_ERROR)
			{
				TC902X_DEBUG_MSG("Transmit error.\n");	
				sp->stats.tx_errors++;
			}
	
			/* Late collision, re-enable transmitter. */
			if (txstatusdword & TC902X_TS_LATE_COLLISION)
			{
				TC902X_DEBUG_MSG("Late collision on transmit.\n");	
				TC902X_WRITE_MACCTRL(baseaddr,
				  TC902X_READ_MACCTRL(baseaddr) |
				  TC902X_MC_TX_ENABLE);
			}
	
			/* Maximum collisions, re-enable transmitter. */
			if (txstatusdword & TC902X_TS_TX_MAX_COLL)
			{
				TC902X_DEBUG_MSG("Maximum collisions on transmit.\n");	

				TC902X_WRITE_MACCTRL(baseaddr,
				  TC902X_READ_MACCTRL(baseaddr) |
				  TC902X_MC_TX_ENABLE);
			}
	
			/* Transmit underrun, reset and re-enable
			 * transmitter.
			 */
			if (txstatusdword & TC902X_TS_TX_UNDERRUN)
			{
				TC902X_DEBUG_MSG("Transmitter underrun.\n");	
				sp->stats.tx_fifo_errors++;
				tc902x_reset(baseaddr, TC902X_AC_TX_RESET |
				                       TC902X_AC_DMA |
				                       TC902X_AC_NETWORK);

				/* Re-configure after DMA reset. */
				if ((tc902x_io_config(tc902x_ethernet_device) < 0) ||
				    (init_tfdlist(tc902x_ethernet_device) < 0))
				{
					printk(KERN_INFO "%s: Error during re-configuration.\n",
				               tc902x_ethernet_device->name);
				}

				TC902X_WRITE_MACCTRL(baseaddr,
				  TC902X_READ_MACCTRL(baseaddr) |
				  TC902X_MC_TX_ENABLE);
			}
		}
		else
			break;

		maxtfdcount--;

	}
	while(maxtfdcount != 0);

	tc902x_nic_txfree(tc902x_ethernet_device);

	return;
}

void tc902x_nic_txfree(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
        /* Free all transmit buffers which have already been transfered
         * via DMA to the TC902X.
         */
 
        int                             NextToFree;
        int                             maxtfdcount;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;
 
	TC902X_DEBUG_MSG("_nic_txfree\n");

	maxtfdcount = TC902X_TFDLIST_LENGTH;

	/* Set the CurrentTxFrameID to skip the next
	 * TxDMACompleteInterrupt.
	 */
	sp->CurrentTxFrameID = 1;

	do
	{
		/* Calculate next TFD to release. */
		NextToFree = (sp->LastFreedTxBuff + 1) % TC902X_TFDLIST_LENGTH;

		TC902X_DEBUG_MSG("TFC = %16.16lx\n", (unsigned long int)
		                 sp->TFDList[NextToFree].TFC);

		/* Look at each TFD's TFC field beginning
		 * at the last freed TFD up to the current TFD.
		 * If the TFDDone bit is set, free the associated
		 * buffer.
		 */
		if ((le64_to_cpu(sp->TFDList[NextToFree].TFC) &
		     TC902X_TFC_TFDDONE) &&
		    (NextToFree != sp->CurrentTFD))
		{
			/* Free the transmit buffer. */
			if (sp->TxBuff[NextToFree] != NULL)
			{
#ifdef TC902X_LINUX2_4
				pci_unmap_single(sp->tc902x_pci_device,
		                 sp->TxBuffDMAhandle[NextToFree].dmahandle,
				 sp->TxBuffDMAhandle[NextToFree].len,
				 PCI_DMA_TODEVICE);
#endif

				TC902X_DEV_KFREE_SKB(sp->TxBuff[NextToFree]);

				sp->TxBuff[NextToFree] = NULL;
			}

			sp->LastFreedTxBuff = NextToFree;
		}
                else
                        break;
 
                maxtfdcount--;
 
        } while(maxtfdcount != 0);
 
	return;
}

int	tc902x_nic_open(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* The TC902X NIC Ethernet interface is opened when activated
	 * by ifconfig.
	 */

	int				phyaddr = 0;
	int				error = 0;
	int				i;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

        TC902X_DEBUG_MSG("_nic_open\n");

#ifdef USE_IO_OPS

	/* Check for I/O port address conflicts, return error
	 * if requested range is already in use.
	 */
	if ((error = check_region(TC902X_PCI_RESOURCE_START(
	                          sp->tc902x_pci_device, 0) &
	                          TC902X_PIB_IOBASEADDRESS,
	                          TC902X_IO_REG_RANGE)) < 0)
	{
		printk(KERN_INFO "%s: Error registering I/O ports.\n",
		       tc902x_ethernet_device->name);
		return error;
	}

	/* Requested range is free, reserve that range for
	 * use by the TC902X.
	 */
	request_region(TC902X_PCI_RESOURCE_START(sp->tc902x_pci_device, 0) &
	               TC902X_PIB_IOBASEADDRESS,
	               TC902X_IO_REG_RANGE,
	               TC902X_DRIVER_NAME);

	/* Use I/O space to access TC902X registers. */
	tc902x_ethernet_device->base_addr =
		(TC902X_PCI_RESOURCE_START(sp->tc902x_pci_device, 0) &
		 TC902X_PIB_IOBASEADDRESS);

#else /* Not using I/O space. */

	/* Use memory space to access TC902X registers. ioremap
	 * is used to map high-memory PCI buffer address space
	 * to user address space.
	 */
	tc902x_ethernet_device->base_addr = (unsigned long)
	  (ioremap(TC902X_PCI_RESOURCE_START(sp->tc902x_pci_device, 1) &
	  TC902X_PMB_MEMBASEADDRESS, TC902X_MEM_REG_RANGE));

#ifdef TC902X_LINUX2_4
	tc902x_ethernet_device->mem_start = pci_resource_start(
	                                     sp->tc902x_pci_device, 1);
	tc902x_ethernet_device->mem_end = tc902x_ethernet_device->mem_start +
	                                   TC902X_MEM_REG_RANGE;
#endif

#endif

	baseaddr = tc902x_ethernet_device->base_addr;

	/* Reset all functions within the TC902X. Do not assert
	 * RST_OUT as not compatible with some PHYs.
	 */
	i = TC902X_AC_GLOBAL_RESET | TC902X_AC_RX_RESET |
	    TC902X_AC_TX_RESET | TC902X_AC_DMA |
	    TC902X_AC_FIFO | TC902X_AC_NETWORK |
	    TC902X_AC_HOST | TC902X_AC_AUTO_INIT;

	error = tc902x_reset(baseaddr, i);
	if (error < 0)
	{
		return error;
	}

	/* Reset PHY. */
	phyaddr = tc902x_find_phyaddr(tc902x_ethernet_device);

	if (phyaddr != -1)
	{
		write_phy_register(tc902x_ethernet_device,
		                   phyaddr, GMII_PHY_CONTROL,
		                   GMII_PHY_CONTROL_RESET);
	}

        /* Intitalize lock variable before requesting interrupt. */
        sp->lock = (spinlock_t) SPIN_LOCK_UNLOCKED;

	/* Check for interrupt line conflicts, and request interrupt
	 * line for TC902X.
	 *
	 * IMPORTANT: Disable TC902X interrupts prior to registering
	 *            IRQ.
	 */
	TC902X_WRITE_INTENABLE(baseaddr, 0x0000);

	/* Register the interrupt line to be used by the TC902X within
	 * the Linux system.
	 */
	if ((error = request_irq(sp->tc902x_pci_device->irq,
	                         &tc902x_interrupt_handler,
	                         SA_SHIRQ,
	                         tc902x_ethernet_device->name,
	                         tc902x_ethernet_device)) < 0)
	{
		printk(KERN_INFO "%s: Error when requesting interrupt.\n",
                               tc902x_ethernet_device->name);
		return error;
	}

	tc902x_ethernet_device->irq = sp->tc902x_pci_device->irq;

#ifdef TC902X_LINUX2_2
	/* Reserve memory for RFD list which must lie on an 8 byte
	 * boundary. So, allocated memory must be adjusted to account
	 * for this alignment.
	 */
	sp->RFDList = kmalloc((sizeof(struct RFD) * TC902X_RFDLIST_LENGTH) +
		              TC902X_DMALIST_ALIGN_PAD, GFP_KERNEL);

	if (sp->RFDList != NULL)
	{
		/* Adjust the start address of the allocated
		 * memory to assure it begins on an 8 byte
		 * boundary.
		 */
		sp->RFDList = (void *)(((long)(sp->RFDList) +
	                       TC902X_DMALIST_ALIGN_PAD) &
		               ~TC902X_DMALIST_ALIGN_PAD);
	}

	/* Reserve memory for TFD list which must lie on an 8 byte
	 * boundary. So, allocated memory must be adjusted to account
	 * for this alignment.
	 */
	sp->TFDList = kmalloc((sizeof(struct TFD) * TC902X_TFDLIST_LENGTH) +
		              TC902X_DMALIST_ALIGN_PAD, GFP_KERNEL);

	if (sp->TFDList != NULL)
	{
		/* Adjust the start address of the allocated
		 * memory to assure it begins on an 8 byte
		 * boundary.
		 */
		sp->TFDList = (void *)(((long)(sp->TFDList) +
	                       TC902X_DMALIST_ALIGN_PAD) &
		               ~TC902X_DMALIST_ALIGN_PAD);
	}
#endif

#ifdef TC902X_LINUX2_4
	sp->RFDList = pci_alloc_consistent(sp->tc902x_pci_device,
	                                   (sizeof(struct RFD) *
	                                   TC902X_RFDLIST_LENGTH),
	                                   &sp->RFDListDMAhandle);

	sp->TFDList = pci_alloc_consistent(sp->tc902x_pci_device,
	                                   (sizeof(struct TFD) *
	                                   TC902X_TFDLIST_LENGTH),
	                                   &sp->TFDListDMAhandle);
#endif

	if ((sp->RFDList == NULL) || (sp->TFDList == NULL))
	{
		printk(KERN_INFO "%s: No memory available for TC9020/9021 RFD and/or TFD lists.\n", tc902x_ethernet_device->name);
		return -ENOMEM;
	}

	error = init_rfdlist(tc902x_ethernet_device);
	if (error < 0)
	{
		printk(KERN_INFO "%s: Error during configuration.\n",
                               tc902x_ethernet_device->name);
		return error;
	}

	error = init_tfdlist(tc902x_ethernet_device);
	if (error < 0)
	{
		printk(KERN_INFO "%s: Error during configuration.\n",
                               tc902x_ethernet_device->name);
		return error;
	}

	/* Set station address in ethernet_device structure. */
	tc902x_ethernet_device->dev_addr[0] =
	  TC902X_READ_STATIONADDRESS0(baseaddr) & 0x00FF;
	tc902x_ethernet_device->dev_addr[1] =
	  (TC902X_READ_STATIONADDRESS0(baseaddr) & 0xFF00) >> 8;
	tc902x_ethernet_device->dev_addr[2] =
	  TC902X_READ_STATIONADDRESS1(baseaddr) & 0x00FF;
	tc902x_ethernet_device->dev_addr[3] =
	  (TC902X_READ_STATIONADDRESS1(baseaddr) & 0xFF00) >> 8;
	tc902x_ethernet_device->dev_addr[4] =
	  TC902X_READ_STATIONADDRESS2(baseaddr) & 0x00FF;
	tc902x_ethernet_device->dev_addr[5] =
	  (TC902X_READ_STATIONADDRESS2(baseaddr) & 0xFF00) >> 8;

	/* Configure TC902X I/O registers. */
	error = tc902x_io_config(tc902x_ethernet_device);
	if (error < 0)
	{
		printk(KERN_INFO "%s: Error during configuration.\n",
                               tc902x_ethernet_device->name);
		return error;
	}

	/* Resolve autonegotiation. */
	if (tc902x_config_autoneg(tc902x_ethernet_device) < 0)
	{
		printk(KERN_INFO "%s: Auto-negotiation error.\n",
		       tc902x_ethernet_device->name);
	}

	/* Enable transmit and receive operation of the TC902X. */
	TC902X_WRITE_MACCTRL(baseaddr, TC902X_READ_MACCTRL(baseaddr) | 
	                     TC902X_MC_RX_ENABLE | TC902X_MC_TX_ENABLE);

#ifdef TC902X_LINUX2_2
	tc902x_ethernet_device->interrupt = 0;
	tc902x_ethernet_device->start = 1;
	clear_bit(0,(void*)&tc902x_ethernet_device->tbusy);
#endif

#ifdef TC902X_LINUX2_4
	netif_start_queue(tc902x_ethernet_device);
#endif

	/* Increment the usage count for this TC902X driver module. */
	MOD_INC_USE_COUNT;

	return 0;
}

int init_rfdlist(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Initialize the RFDList. */

	int				i;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_init_rfdlist\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	/* Clear the receive buffer not ready flag. */
	sp->RxBuffNotReady = 0;

	for(i=0; i<TC902X_RFDLIST_LENGTH; i++)
	{
		/* Free any allocated receive buffers. */
#ifdef TC902X_LINUX2_4
		pci_unmap_single(sp->tc902x_pci_device,
                                 sp->RxBuffDMAhandle[i].dmahandle,
		                 sp->RxBuffDMAhandle[i].len,
		                 PCI_DMA_FROMDEVICE);
#endif
		if (sp->RxBuff[i] != NULL)
			TC902X_DEV_KFREE_SKB(sp->RxBuff[i]);
		sp->RxBuff[i] = NULL;

		/* Clear out the RFS field. */
	 	sp->RFDList[i].RFS = 0x0000000000000000; 

		if (tc902x_get_rxbuff(tc902x_ethernet_device, i) < 0) 
		{
			/* A receive buffer was not ready, break the
			 * RFD list here and set the receive buffer
			 * not ready flag.
			 */
			sp->RxBuffNotReady = 1;

			TC902X_DEBUG_MSG("Cannot allocate Rx buffer.\n");

			/* Just in case we cannot allocate a single RFD.
			 * Should not occur.
			 */
			if (i == 0)
			{
				printk(KERN_ERR "%s: No memory available for RFD list.\n",
				       tc902x_ethernet_device->name);
				return -ENOMEM;
			}
		}

		/* Set up RFDs to point to each other. A ring structure. */
#ifdef TC902X_LINUX2_2
	 	sp->RFDList[i].RFDNextPtr = cpu_to_le64(
		                                 TC902X_HOST2BUS_MAP(
		                                 &sp->RFDList[(i + 1) %
		                                 TC902X_RFDLIST_LENGTH]));
#endif

#ifdef TC902X_LINUX2_4
	 	sp->RFDList[i].RFDNextPtr = cpu_to_le64(
		                            sp->RFDListDMAhandle +
		                            ((sizeof(struct RFD)) *
		                             ((i + 1) %
		                              TC902X_RFDLIST_LENGTH)));
#endif
	}

	sp->CurrentRFD = 0;
	sp->LastRestoredRxBuff = i - 1;

	/* Write the location of the RFDList to the TC902X. */
#ifdef TC902X_LINUX2_2
	TC902X_WRITE_RFDLISTPTR0(baseaddr,
	  (u32)(TC902X_HOST2BUS_MAP(sp->RFDList)));
	TC902X_WRITE_RFDLISTPTR1(baseaddr, 0x00000000);
#endif

#ifdef TC902X_LINUX2_4
	TC902X_WRITE_RFDLISTPTR0(baseaddr, (u32)(sp->RFDListDMAhandle));
	TC902X_WRITE_RFDLISTPTR1(baseaddr, 0x00000000);
#endif

	return 0;
}

int init_tfdlist(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Initialize TFDList. */

	int				i;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_init_tfdlist\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	for(i=0; i<TC902X_TFDLIST_LENGTH; i++)
	{
#ifdef TC902X_LINUX2_2
	 	sp->TFDList[i].TFDNextPtr = cpu_to_le64(
		                                 TC902X_HOST2BUS_MAP(
		                                 &sp->TFDList[(i + 1) %
		                                 TC902X_TFDLIST_LENGTH]));
#endif

#ifdef TC902X_LINUX2_4
	 	sp->TFDList[i].TFDNextPtr = cpu_to_le64(
		                            sp->TFDListDMAhandle +
		                            ((sizeof(struct TFD)) *
		                             ((i + 1) %
		                              TC902X_TFDLIST_LENGTH)));
#endif

	 	sp->TFDList[i].TFC = cpu_to_le64(TC902X_TFC_TFDDONE);
		if (sp->TxBuff[i] != NULL)
			TC902X_DEV_KFREE_SKB(sp->TxBuff[i]);
		sp->TxBuff[i] = NULL;
	}

	sp->CurrentTFD = TC902X_TFDLIST_LENGTH - 1;
	sp->CurrentTxFrameID = 0;
	sp->LastFreedTxBuff = TC902X_TFDLIST_LENGTH - 1;

	/* Write the location of the TFDList to the TC902X. */
#ifdef TC902X_LINUX2_2
	TC902X_WRITE_TFDLISTPTR0(baseaddr, (u32)(TC902X_HOST2BUS_MAP(
	                                         sp->TFDList)));
	TC902X_WRITE_TFDLISTPTR1(baseaddr, 0x00000000);
#endif

#ifdef TC902X_LINUX2_4
	TC902X_DDEBUG_MSG("Starting TFDListPtr = %8.8x\n",
	  (u32)(sp->TFDListDMAhandle));
	TC902X_WRITE_TFDLISTPTR0(baseaddr, (u32)(sp->TFDListDMAhandle));
	TC902X_WRITE_TFDLISTPTR1(baseaddr, 0x00000000);
#endif

	return 0;
}

int	tc902x_get_rxbuff(TC902X_DEVICE_TYPE *tc902x_ethernet_device, int rfd)
{
	/* Create a receive buffer within system memory and update
	 * NIC private structure appropriately.
	 */
	u64				rxfragsize;
	struct sk_buff			*skb;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_get_rxbuff\n");

	/* Allocate memory buffers for receive frames. Pad by
	 * 2 to account for IP field alignment.
	 */
	skb = dev_alloc_skb(TC902X_RXFRAG_SIZE + 2);

	if (skb == NULL)
	{
		sp->RxBuff[rfd] = NULL;
		return -ENOMEM;
	}

	/* Adjust the data start location within the buffer to
	 * align IP address field to a 16 byte boundary.
	 */
	skb_reserve(skb, 2);

	/* Associate the receive buffer with the TC902X NIC. */
	skb->dev = tc902x_ethernet_device;

	/* Save the address of the sk_buff structure. */
	sp->RxBuff[rfd] = skb;

#ifdef TC902X_LINUX2_2
	/* The sk_buff struct "data" field holds the address
	 * of the beginning of valid octets. Assign this address
	 * to the fragement address field of the RFD.
	 */
 	sp->RFDList[rfd].FragInfo = cpu_to_le64(
	                                 TC902X_HOST2BUS_MAP(skb->data));
#endif

#ifdef TC902X_LINUX2_4
	sp->RxBuffDMAhandle[rfd].len = TC902X_RXFRAG_SIZE;
	sp->RxBuffDMAhandle[rfd].dmahandle = pci_map_single(
	                                      sp->tc902x_pci_device,
	                                      skb->data, TC902X_RXFRAG_SIZE,
	                                      PCI_DMA_FROMDEVICE);
	sp->RFDList[rfd].FragInfo = cpu_to_le64(
	                                sp->RxBuffDMAhandle[rfd].dmahandle);
#endif

	/* Set the RFD fragment length. */
	rxfragsize = TC902X_RXFRAG_SIZE;
 	sp->RFDList[rfd].FragInfo |= cpu_to_le64((rxfragsize << 48) &
	                                         TC902X_RFI_FRAGLEN);

	return 0;
}

int	tc902x_nic_stop(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Release resources requested by driver open function. */

	int	i;
	int	error;
	u32	baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_nic_stop\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	TC902X_DDEBUG_MSG("TFDunavailCount = %i\n", sp->TFDunavailCount);
	TC902X_DDEBUG_MSG("RFDlistendCount = %i\n", sp->RFDlistendCount);
	TC902X_DDEBUG_MSG("RFDListCheckedCount = %i\n", sp->RFDListCheckedCount);
	TC902X_DDEBUG_MSG("EmptyRFDListCount = %i\n", sp->EmptyRFDListCount);
	TC902X_DUMPTFDLIST(tc902x_ethernet_device);

	/* Reset all functions within the TC902X to shut down the
	 * TX902X.
	 */
	i = TC902X_AC_GLOBAL_RESET | TC902X_AC_RX_RESET |
	    TC902X_AC_TX_RESET | TC902X_AC_DMA |
	    TC902X_AC_FIFO | TC902X_AC_NETWORK |
	    TC902X_AC_HOST | TC902X_AC_AUTO_INIT;

	error = tc902x_reset(baseaddr, i);
	if (error < 0)
	{
		return error;
	}

	/* Decrement the usage count for this TC902X driver module. */
	MOD_DEC_USE_COUNT;

	/* Free all receive buffers. */
	for(i=0; i<TC902X_RFDLIST_LENGTH; i++)
	{
#ifdef TC902X_LINUX2_4
		pci_unmap_single(sp->tc902x_pci_device,
                                 sp->RxBuffDMAhandle[i].dmahandle,
		                 sp->RxBuffDMAhandle[i].len,
		                 PCI_DMA_FROMDEVICE);
#endif
		if (sp->RxBuff[i] != NULL)
			TC902X_DEV_KFREE_SKB(sp->RxBuff[i]);
		sp->RxBuff[i] = NULL;
	}

	/* Free all transmit buffers. */
	for(i=0; i<TC902X_TFDLIST_LENGTH; i++)
	{
		if (sp->TxBuff[i] != NULL)
			TC902X_DEV_KFREE_SKB(sp->TxBuff[i]);
		sp->TxBuff[i] = NULL;
	}

#ifdef TC902X_LINUX2_2
	tc902x_ethernet_device->start = 0;
	set_bit(0,(void*)&tc902x_ethernet_device->tbusy);

	/* Free memory associated with the RFDList. */
	kfree(sp->RFDList);

	/* Free memory associated with the TFDList. */
	kfree(sp->TFDList);
#endif

#ifdef TC902X_LINUX2_4
	netif_stop_queue(tc902x_ethernet_device);

	/* Free memory associated with the RFDList. */
	pci_free_consistent(sp->tc902x_pci_device,
                            (sizeof(struct RFD) *
                            TC902X_RFDLIST_LENGTH),
	                    sp->RFDList, sp->RFDListDMAhandle);
		
	/* Free memory associated with the TFDList. */
	pci_free_consistent(sp->tc902x_pci_device,
                            (sizeof(struct TFD) *
                            TC902X_TFDLIST_LENGTH),
	                    sp->TFDList, sp->TFDListDMAhandle);
#endif

	/* Release interrupt line. */
	free_irq(tc902x_ethernet_device->irq, tc902x_ethernet_device);

#ifdef USE_IO_OPS
	
	/* Release I/O range reserved for TC902X registers. */
	release_region(tc902x_ethernet_device->base_addr,
	               TC902X_IO_REG_RANGE);
	
#else /* Not using I/O space. */
	
	/* Unmap memory used for TC902X registers. */

	/* The following line produces strange results unless
	 * unregister_netdev precedes it.
	 */
	iounmap((void *)tc902x_ethernet_device->base_addr);
	
#endif

	return 0;
}

int	tc902x_nic_hard_start_xmit(struct sk_buff *skb,
	                          TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Transmit an Ethernet frame. */

	u64				fraglen;
	u64				vlanvid;
	u64				vlancfi;
	u64				vlanuserpriority;
        unsigned long                   flags;
	int				NextTFD;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

        TC902X_DDEBUG_MSG("_nic_hard_start_xmit\n");

	baseaddr = tc902x_ethernet_device->base_addr;

#ifdef TC902X_LINUX2_2
	/* If TC902X NIC is already busy, return error. */
	if (test_and_set_bit(0, (void*)&tc902x_ethernet_device->tbusy) != 0)
	{
                TC902X_DEBUG_MSG("Transmit busy!\n");

		return -EBUSY;
	}
#endif

        /* Disable interrupts. */
        spin_lock_irqsave(&sp->lock, flags);

#ifdef TC902X_LINUX2_4
	/* If in 10Mbps mode, stop the transmit queue so
	 * no more transmit frames are accepted.
	 */
	if (sp->tenmbpsmode)
	{
		netif_stop_queue(tc902x_ethernet_device);
	}
#endif

	/* Next TFD is found by incrementing the CurrentTFD
	 * counter, modulus the length of the TFDList.
	 */
	NextTFD = (sp->CurrentTFD + 1) % TC902X_TFDLIST_LENGTH;

	/* Check for availability of next TFD. */
	if (!(le64_to_cpu(sp->TFDList[NextTFD].TFC) &
	      TC902X_TFC_TFDDONE) || (NextTFD == sp->LastFreedTxBuff))
	{
		TC902X_DEBUG_MSG("Next TFD not available.\n");

		/* Attempt to free any used TFDs. */
		tc902x_nic_txfree(tc902x_ethernet_device);

#ifdef TC902X_LINUX2_2
                /* Transmit no longer busy. */
		clear_bit(0,(void*)&tc902x_ethernet_device->tbusy);
#endif
 
                /* Restore interrupts. */
                spin_unlock_irqrestore(&sp->lock, flags);

#ifdef TC902X_DEBUG
		/* Increment the TFDunavailCount counter. */
		sp->TFDunavailCount++;
#endif
 
                return -ENOMEM;
	}

#ifdef TC902X_LINUX2_4
	sp->TxBuffDMAhandle[NextTFD].len = skb->len;
	sp->TxBuffDMAhandle[NextTFD].dmahandle = pci_map_single(
	                                          sp->tc902x_pci_device,
	                                          skb->data, skb->len,
	                                          PCI_DMA_TODEVICE);
#endif

	/* Save the sk_buff pointer so interrupt handler can later free
	 * memory occupied by buffer.
	 */
	sp->TxBuff[NextTFD] = skb;

	/* Clear all TFC fields, except TFDDONE. */
	sp->TFDList[NextTFD].TFC = cpu_to_le64(TC902X_TFC_TFDDONE);

	/* Specify the TFC field within the TFD. */
	sp->TFDList[NextTFD].
		TFC |= cpu_to_le64(TC902X_TFC_WORDALIGNDISABLED |
		                   (TC902X_TFC_FRAMEID &
		                    cpu_to_le64(sp->CurrentTxFrameID)) |
		                   (TC902X_TFC_FRAGCOUNT & (1 << 24)));

	/* Request TxComplete interrupts at an interval defined
	 * by the constant TC902X_FRAMESBETWEENTXCOMPLETES.
	 * Request TxComplete interrupt for every frame
	 * if in 10Mbps mode to accomodate problem with 10Mbps
	 * processing.
	 */
	if (sp->tenmbpsmode)
	{
		sp->TFDList[NextTFD].TFC |=
	        cpu_to_le64(TC902X_TFC_TXINDICATE);
	}
	else if ((sp->CurrentTxFrameID %
	         TC902X_FRAMESBETWEENTXDMACOMPLETES) == 0)
	{
		sp->TFDList[NextTFD].TFC |=
	        cpu_to_le64(TC902X_TFC_TXDMAINDICATE);
	}

        /* Based on compilation option, determine if FCS is to be
         * appended to transmit frame by TC902X.
         */
        if (!(TC902X_APPEND_FCS_ON_TX))
        {
		sp->TFDList[NextTFD].
                        TFC |= cpu_to_le64(TC902X_TFC_FCSAPPENDDISABLE);
        }

        /* Based on compilation option, determine if IP, TCP and/or
	 * UDP checksums are to be added to transmit frame by TC902X.
         */
        if (TC902X_ADD_IPCHECKSUM_ON_TX)
        {
		sp->TFDList[NextTFD].
                        TFC |= cpu_to_le64(TC902X_TFC_IPCHECKSUMENABLE);
        }
        if (TC902X_ADD_TCPCHECKSUM_ON_TX)
        {
		sp->TFDList[NextTFD].
                        TFC |= cpu_to_le64(TC902X_TFC_TCPCHECKSUMENABLE);
        }
        if (TC902X_ADD_UDPCHECKSUM_ON_TX)
        {
		sp->TFDList[NextTFD].
                        TFC |= cpu_to_le64(TC902X_TFC_UDPCHECKSUMENABLE);
        }

        /* Based on compilation option, determine if VLAN tag info is to be
         * inserted into transmit frame by TC902X.
         */
        if (TC902X_INSERT_MANUAL_VLAN_TAG)
        {
		vlanvid = TC902X_MANUAL_VLAN_VID;
		vlancfi = TC902X_MANUAL_VLAN_CFI;
		vlanuserpriority = TC902X_MANUAL_VLAN_USERPRIORITY;

		sp->TFDList[NextTFD].TFC |= cpu_to_le64(
		                            TC902X_TFC_VLANTAGINSERT |
                                            (vlanvid << 32) |
			                    (vlancfi << 44) |
			                    (vlanuserpriority << 45));
        }

	/* The fragment start location within system memory is defined
	 * by the sk_buff structure's data field. The physical address
	 * of this location within the system's virtual memory space
	 * is determined using the TC902X_HOST2BUS_MAP function.
	 */
#ifdef TC902X_LINUX2_2
	sp->TFDList[NextTFD].FragInfo = cpu_to_le64(
	                                     TC902X_HOST2BUS_MAP(skb->data));
#endif

#ifdef TC902X_LINUX2_4
	sp->TFDList[NextTFD].FragInfo = cpu_to_le64(
	                                sp->TxBuffDMAhandle[NextTFD].dmahandle);
#endif
	
	/* The length of the fragment within system memory is defined by
	 * the sk_buff structure's len field.
	 */
	fraglen = (u16)skb->len;
	sp->TFDList[NextTFD].FragInfo |= cpu_to_le64(TC902X_TFI_FRAGLEN &
		                         (fraglen << 48));

	/* Clear the TFDDone bit last to indicate the TFD is ready
	 * for transfer to the TC902X.
	 */
	sp->TFDList[NextTFD].TFC &= cpu_to_le64(~TC902X_TFC_TFDDONE);

	/* Record frame transmit start time (jiffies = Linux
	 * kernel current time stamp).
	 */
	tc902x_ethernet_device->trans_start = jiffies;

	/* Update current TFD indicator. */
	sp->CurrentTFD = NextTFD;

	/* Calculate the new ID for the next transmit frame by
	 * incrementing the CurrentTxFrameID counter.
	 */
	sp->CurrentTxFrameID++;

	/* Force a transmit DMA poll event. */
	TC902X_WRITE_DMACTRL(baseaddr, TC902X_DC_TX_DMA_POLL_NOW);

        /* Restore interrupts. */
        spin_unlock_irqrestore(&sp->lock, flags);

#ifdef TC902X_LINUX2_2
        /* Transmit no longer busy.
	 * If in 10Mbps mode, do not indicate transmit is free
	 * until receive TxComplete interrupt.
	 */
	if (!(sp->tenmbpsmode))
	{
		clear_bit(0,(void*)&tc902x_ethernet_device->tbusy);
	}
#endif

	return 0;
}

int tc902x_nic_rx(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Transfer received Ethernet frames to higher network layers. */

	int				maxrfdcount;
	int				framelen;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;
	struct sk_buff			*skb;

	TC902X_DEBUG_MSG("_nic_rx\n");

	baseaddr = tc902x_ethernet_device->base_addr;
	maxrfdcount = TC902X_MAXRFDPROCESS_COUNT;

	while((le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS)) &
	       TC902X_RFS_RFDDONE &&
	      (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS)) &
	       TC902X_RFS_FRAMESTART &&
	      (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS)) &
	       TC902X_RFS_FRAMEEND &&
	      (sp->RxBuff[sp->CurrentRFD] != NULL)) 
	{
                if (--maxrfdcount == 0)
                {
                        /* There are more RFDs to process, however the
			 * allocated amount of RFD processing time has
			 * expired. Assert Interrupt Requested to make
			 * sure we come back to process the remaining RFDs.
                         */
                        TC902X_WRITE_ASICCTRL(baseaddr,
			                      TC902X_READ_ASICCTRL(baseaddr) |
                                              TC902X_AC_INT_REQUEST);
                        break;
                }
 
		/* Get received frame length. */
		framelen = le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS) &
	       		        TC902X_RFS_RXFRAMELEN;

		/* Check for jumbo frame arrival with too small
		 * RXFRAG_SIZE.
		 */
		if (framelen > TC902X_RXFRAG_SIZE)
		{
			TC902X_DEBUG_MSG("RFS FrameLen > allocated fragment size.\n");

			framelen = TC902X_RXFRAG_SIZE;
		}

		/* Get the received frame buffer. */
		skb = sp->RxBuff[sp->CurrentRFD];

		if ((TC902X_DROP_ON_RX_ETH_ERRORS &&
		     (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
		      (TC902X_RFS_RXFIFOOVERRUN |
		       TC902X_RFS_RXRUNTFRAME | 
		       TC902X_RFS_RXALIGNMENTERROR |
		       TC902X_RFS_RXFCSERROR |
		       TC902X_RFS_RXOVERSIZEDFRAME |
		       TC902X_RFS_RXLENGTHERROR)))))
		{

			TC902X_DEBUG_MSG("Rx error, RFS = %16.16lx\n", (unsigned long int) sp->RFDList[sp->CurrentRFD].RFS);

			/* Increment general receive error statistic. */
			sp->stats.rx_errors++;

			/* Increment detailed receive error statistics. */
			if (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			    TC902X_RFS_RXFIFOOVERRUN))
			{
				TC902X_DEBUG_MSG("RX FIFO overrun occured.\n");

				sp->stats.rx_fifo_errors++;

				if (sp->RxBuffNotReady == 1)
				{
					/* If experience a RxFIFO overrun, and
					 * the RxBuffNotReady flag is set,
					 * assume the FIFO overran due to lack
					 * of an RFD.
					 */
					sp->stats.rx_dropped++;
				}
			}

			if (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			    TC902X_RFS_RXRUNTFRAME))
			{
				TC902X_DEBUG_MSG("RX runt occured.\n");
				sp->stats.rx_length_errors++;
			}

			if (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			    TC902X_RFS_RXOVERSIZEDFRAME));
			/* Do nothing, error count handled by a TC902X
			 * statistic register.
			 */

			if (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			    TC902X_RFS_RXALIGNMENTERROR))
			{
				TC902X_DEBUG_MSG("RX alignment error occured.\n");
				sp->stats.rx_frame_errors++;
			}


			if (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			    TC902X_RFS_RXFCSERROR));
			/* Do nothing, error count handled by a TC902X
			 * statistic register.
			 */

			/* Free the memory associated with the RX
			 * buffer since it is erroneous and we will
			 * not pass it to higher layer processes.
			 */
			if (sp->RxBuff[sp->CurrentRFD] != NULL)
			{
#ifdef TC902X_LINUX2_4
				pci_unmap_single(sp->tc902x_pci_device,
                                 sp->RxBuffDMAhandle[sp->CurrentRFD].dmahandle,
		                 sp->RxBuffDMAhandle[sp->CurrentRFD].len,
		                 PCI_DMA_FROMDEVICE);
#endif

				TC902X_DEV_KFREE_SKB(sp->RxBuff[sp->CurrentRFD]);
			}

		}
		else
		{

			/* Adjust the new buffer length to accomodate the size
			 * of the received frame.
			 */
			skb_put(skb, framelen);
	
			/* Set the buffer's protocol field to Ethernet. */
			skb->protocol = eth_type_trans(skb,
			                               tc902x_ethernet_device);

			/* If the frame contains an IP/TCP/UDP frame,
			 * determine if upper layer must check IP/TCP/UDP
			 * checksums.
			 *
			 * NOTE: DO NOT RELY ON THE TCP/UDP CHECKSUM
			 *       VERIFICATION FOR SILICON REVISIONS B3
			 *       AND EARLIER!
			 *
			if ((le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			     (TC902X_RFS_TCPDETECTED |
			      TC902X_RFS_UDPDETECTED |
			      TC902X_RFS_IPDETECTED))) &&
			    !(le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
			      (TC902X_RFS_TCPERROR |
			       TC902X_RFS_UDPERROR |
			       TC902X_RFS_IPERROR))))
			{
				* Indicate IP checksums were performed
				 * by the TC902X.
				 *
				skb->ip_summed = CHECKSUM_UNNECESSARY;
			}
			else
			*/
			if (1==1)
			{
				/* The TC902X encountered an error with (or
				 * there were no) IP/TCP/UDP checksums.
				 * This may or may not indicate an invalid
				 * IP/TCP/UDP frame was received. Let the
				 * upper layer decide.
				 */
				skb->ip_summed = CHECKSUM_NONE;
			}
	
			/* Hand off frame for higher layer processing.
			 * The function netif_rx() releases the sk_buff
			 * when processing completes.
			 */
			netif_rx(skb);

			/* Record frame receive time (jiffies = Linux
			 * kernel current time stamp).
			 */
			tc902x_ethernet_device->last_rx = jiffies;
		}

		/* Assure RX buffer is not reused by TC902X. */
		sp->RxBuff[sp->CurrentRFD] = NULL;

		/* Increment the current RFD counter. */
		sp->CurrentRFD = (sp->CurrentRFD + 1) % TC902X_RFDLIST_LENGTH;
	
	}


#ifdef TC902X_DEBUG
	/* Check if the RFD list contained no receive frame data. */
	if (maxrfdcount == TC902X_MAXRFDPROCESS_COUNT)
	{
		sp->EmptyRFDListCount++;
	}
#endif
	while((le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
	       TC902X_RFS_RFDDONE)) &&
	      !((le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
	       TC902X_RFS_FRAMESTART)) &&
	      (le64_to_cpu(sp->RFDList[sp->CurrentRFD].RFS &
	       TC902X_RFS_FRAMEEND))))
	{
		TC902X_DEBUG_MSG("Frame requires multiple RFDs.\n");

		/* An unexpected event, additional code needed to handle
		 * properly. So for the time being, just disregard the
		 * frame.
		 */

		/* Free the memory associated with the RX
		 * buffer since it is erroneous and we will
		 * not pass it to higher layer processes.
		 */
		if (sp->RxBuff[sp->CurrentRFD] != NULL)
		{
#ifdef TC902X_LINUX2_4
			pci_unmap_single(sp->tc902x_pci_device,
                                sp->RxBuffDMAhandle[sp->CurrentRFD].dmahandle,
		                sp->RxBuffDMAhandle[sp->CurrentRFD].len,
		                PCI_DMA_FROMDEVICE);
#endif
			TC902X_DEV_KFREE_SKB(sp->RxBuff[sp->CurrentRFD]);
		}

		/* Assure RX buffer is not reused by TC902X. */
		sp->RxBuff[sp->CurrentRFD] = NULL;

		/* Increment the current RFD counter. */
		sp->CurrentRFD = (sp->CurrentRFD + 1) % TC902X_RFDLIST_LENGTH;
	}

        /* Check to see if there are a minimum number of used
         * RFDs before restoring any (should improve performance.)
         */
        if (((sp->CurrentRFD > sp->LastRestoredRxBuff) &&
             ((sp->LastRestoredRxBuff + TC902X_MINUSEDRFDSTOFREE) <=
              sp->CurrentRFD)) ||
            ((sp->CurrentRFD < sp->LastRestoredRxBuff) &&
             ((sp->LastRestoredRxBuff + TC902X_MINUSEDRFDSTOFREE) <=
              (sp->CurrentRFD + TC902X_RFDLIST_LENGTH))))
        {
		tc902x_nic_rxrestore(tc902x_ethernet_device);
	}

	return 0;
}

int tc902x_nic_rxrestore(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Restore used receive buffers. */
 
	int                             i;
	struct tc902x_nic_private        *sp = (struct tc902x_nic_private *)
					tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_nic_rxrestore\n");

	/* Assume receive buffers will be available. */
	sp->RxBuffNotReady = 0;

        while (sp->RxBuff[i = ((sp->LastRestoredRxBuff + 1) %
               TC902X_RFDLIST_LENGTH)] == NULL)
	{
		/* Generate a new receive buffer to replace the
		 * current buffer (which will be released by the
		 * Linux system).
		 */
		if (tc902x_get_rxbuff(tc902x_ethernet_device, i) < 0)
		{
			TC902X_DEBUG_MSG("Cannot allocate new Rx buffer.\n");

			/* Mark a flag indicating a receive buffer
			 * was not available. Use this flag to update
			 * the rx_dropped Linux statistic.
			 */
			sp->RxBuffNotReady = 1;

			break;
		}

		/* Reset the RFS field. */
		sp->RFDList[i].RFS = 0x0000000000000000;

		sp->LastRestoredRxBuff = i;
	}
	
	return 0;
}

TC902X_STATS_TYPE* tc902x_nic_get_stats(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Provides statistical information about the TC902X NIC. */

	u16				temp1;
	u16				temp2;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_nic_get_stats\n");

	/* Check to see if the NIC has been initialized via nic_open,
	 * before trying to read statistic registers.
	 */
#ifdef TC902X_LINUX2_2
	if (tc902x_ethernet_device->start == 0)
	{
#endif
#ifdef TC902X_LINUX2_4
	if (!test_bit(__LINK_STATE_START, &tc902x_ethernet_device->state))
	{
#endif
		return &sp->stats;
	}


	baseaddr = tc902x_ethernet_device->base_addr;

        sp->stats.rx_packets += TC902X_READ_FRAMESRCVDOK(baseaddr);
        sp->stats.tx_packets += TC902X_READ_FRAMESXMTDOK(baseaddr);
        sp->stats.rx_bytes   += TC902X_READ_OCTETRCVOK(baseaddr);
        sp->stats.tx_bytes   += TC902X_READ_OCTETXMTOK(baseaddr);
	temp1 = TC902X_READ_FRAMESLOSTRXERRORS(baseaddr);
        sp->stats.rx_errors  += temp1;
        sp->stats.rx_missed_errors += temp1;
	temp1 = TC902X_READ_SINGLECOLFRAMES(baseaddr) +
	        TC902X_READ_MULTICOLFRAMES(baseaddr) +
	        TC902X_READ_LATECOLLISIONS(baseaddr);
	temp2 = TC902X_READ_CARRIERSENSEERRORS(baseaddr);
        sp->stats.collisions += temp1;
	sp->stats.tx_dropped += TC902X_READ_FRAMESABORTXSCOLLS(baseaddr);
        sp->stats.tx_errors  += TC902X_READ_FRAMESWEXDEFERRAL(baseaddr) +
	                        TC902X_READ_FRAMESWDEFERREDXMT(baseaddr) +
	                        temp1 + temp2;
        sp->stats.multicast  += TC902X_READ_MCSTOCTETRCVDOK(baseaddr);

        /* detailed tx_errors */
        sp->stats.tx_carrier_errors += temp2;

        /* detailed rx_errors */
	sp->stats.rx_length_errors += TC902X_READ_INRANGELENGTHERRORS(baseaddr) +
	                              TC902X_READ_FRAMETOOLONGERRRORS(baseaddr);
	sp->stats.rx_crc_errors += TC902X_READ_FRAMECHECKSEQERRORS(baseaddr);

	/* Unutilized TC902X statistic registers. */
	TC902X_READ_MCSTFRAMESRCVDOK(baseaddr);

        /* Masked TC902X statistic registers (need not be read to clear)
	TC902X_READ_MACCONTROLFRAMESXMTDOK(baseaddr);
	TC902X_READ_BCSTFRAMESXMTDOK(baseaddr);
	TC902X_READ_MCSTFRAMESXMTDOK(baseaddr);
	TC902X_READ_BCSTOCTETXMTOK(baseaddr);
	TC902X_READ_MCSTOCTETXMTOK(baseaddr);
	TC902X_READ_MACCONTROLFRAMESRCVD(baseaddr);
	TC902X_READ_BCSTFRAMESRCVDOK(baseaddr);
	TC902X_READ_BCSTOCTETRCVOK(baseaddr);
	TC902X_READ_TXJUMBOFRAMES(baseaddr);
	TC902X_READ_UDPCHECKSUMERRORS(baseaddr);
	TC902X_READ_IPCHECKSUMERRORS(baseaddr);
	TC902X_READ_TCPCHECKSUMERRORS(baseaddr);
	TC902X_READ_RXJUMBOFRAMES(baseaddr);
	*/


	/* Unutilized RMON statistic registers. */

        /* Masked TC902X statistic registers (need not be read to clear)
	TC902X_READ_ETHERSTATSCOLLISIONS(baseaddr);
	TC902X_READ_ETHERSTATSOCTETSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS64OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS65TO127OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS128TO255OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS256TO511OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS512TO1023OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSPKTS1024TO1518OCTESTSTRANSMIT(baseaddr);
	TC902X_READ_ETHERSTATSCRCALIGNERRORS(baseaddr);
	TC902X_READ_ETHERSTATSUNDERSIZEPKTS(baseaddr);
	TC902X_READ_ETHERSTATSFRAGMENTS(baseaddr);
	TC902X_READ_ETHERSTATSJABBERS(baseaddr);
	TC902X_READ_ETHERSTATSOCTETS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS64OCTESTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS65TO127OCTESTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS128TO255OCTESTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS256TO511OCTESTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS512TO1023OCTESTS(baseaddr);
	TC902X_READ_ETHERSTATSPKTS1024TO1518OCTESTS(baseaddr);
	*/

	return &sp->stats;
}

void tc902x_nic_set_multicast_list(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Determine and configure multicast operation and set
	 * receive mode for TC902X.
	 */
	u8			receivemode;
	u32			hashtable[2];
	unsigned int		hashindex;
	u32			baseaddr;
	struct dev_mc_list	*mc_list_ptr;


	TC902X_DEBUG_MSG("_nic_set_multicast_list\n");

	baseaddr = tc902x_ethernet_device->base_addr;

	receivemode = TC902X_RM_RECEIVEUNICAST |
	              TC902X_RM_RECEIVEBROADCAST;

	if (tc902x_ethernet_device->flags & IFF_PROMISC)
	{
		/* NIC to be configured in promiscuous mode. */
		receivemode = TC902X_RM_RECEIVEALLFRAMES;
	}
	else if ((tc902x_ethernet_device->flags & IFF_ALLMULTI) ||
	         (tc902x_ethernet_device->flags & IFF_MULTICAST &
	          (tc902x_ethernet_device->mc_count >
	           TC902X_MULTICAST_HASHTABLE_SIZE)))
	{
		/* NIC to be configured to receive all multicast
		 * frames. */
		receivemode |= TC902X_RM_RECEIVEMULTICAST;
	}
	else if (tc902x_ethernet_device->flags & IFF_MULTICAST &
	    (tc902x_ethernet_device->mc_count > 0))
	{
		/* NIC to be configured to receive selected
		 * multicast addresses. */
		receivemode |= TC902X_RM_RECEIVEMULTICASTHASH;
	}

	/* Calculate the bits to set for the 64 bit, TC902X HASHTABLE.
	 * The TC902X applies a cyclic-redundancy-check (the same CRC
	 * used to calculate the frame data FCS) to the destination
	 * address all incoming multicast frames whose destination
	 * address has the multicast bit set. The least significant
	 * 6 bits of the CRC result are used as an addressing index
	 * into the hash table. If the value of the bit addressed by
	 * this index is a 1, the frame is passed to the host system.
	 */

	/* Clear hashtable. */
	hashtable[0] = 0x00000000;
	hashtable[1] = 0x00000000;

	/* Cycle through all multicast addresses to filter.*/
	for (mc_list_ptr = tc902x_ethernet_device->mc_list;
	     mc_list_ptr != NULL;
	     mc_list_ptr = mc_list_ptr->next)
	{
		/* Calculate CRC result for each multicast address. */
		hashindex = ether_crc_le(ETH_ALEN, mc_list_ptr->dmi_addr);

		/* Use only the least significant 6 bits. */
		hashindex = hashindex & 0x3F;

		/* Within "hashtable", set bit number "hashindex"
		 * to a logic 1.
		 */
		set_bit(hashindex, hashtable);
	}

	/* Write the value of the hashtable, to the 4, 16 bit
	 * HASHTABLE TC902X registers.
	 */
	TC902X_WRITE_HASHTABLE0(baseaddr, hashtable[0]);
	TC902X_WRITE_HASHTABLE1(baseaddr, hashtable[1]);

	TC902X_WRITE_RECEIVEMODE(baseaddr, receivemode);

	TC902X_DEBUG_MSG("ReceiveMode = %x\n",TC902X_READ_RECEIVEMODE(baseaddr));

	return;
}

/*
 * The following code fragment was authored by Donald Becker.
 */

/* The little-endian AUTODIN II ethernet CRC calculations.
   A big-endian version is also available.
   This is slow but compact code.  Do not use this routine for bulk data,
   use a table-based routine instead.
   This is common code and should be moved to net/core/crc.c.
   Chips may use the upper or lower CRC bits, and may reverse and/or invert
   them.  Select the endian-ness that results in minimal calculations.
*/
unsigned const ethernet_polynomial_le = 0xedb88320U;
unsigned ether_crc_le(int length, unsigned char *data)
{
        unsigned int crc = 0xffffffff;  /* Initial value. */
        while(--length >= 0) {
                unsigned char current_octet = *data++;
                int bit;
                for (bit = 8; --bit >= 0; current_octet >>= 1) {
                        if ((crc ^ current_octet) & 1) {
                                crc >>= 1;
                                crc ^= ethernet_polynomial_le;
                        } else
                                crc >>= 1;
                }
        }
        return crc;
}
/*
 * End of code fragment authored by Donald Becker.
 */

int tc902x_nic_init(TC902X_DEVICE_TYPE *tc902x_ethernet_device)
{
	/* Initialize TC902X NIC. */

	struct tc902x_nic_private	*sp = NULL;

	TC902X_DEBUG_MSG("_nic_init\n");

	/* Register the TC902X NIC in the list of Ethernet devices. */
	tc902x_ethernet_device = init_etherdev(tc902x_ethernet_device,
	                         sizeof(struct tc902x_nic_private));

	if (tc902x_ethernet_device == NULL)
	{
		printk(KERN_INFO "Could not initialize TC9020/9021 based NIC.\n");
		return -ENODEV;
	}

	/* Reserve memory for tc902x_nic_private structure. */
	sp = kmalloc(sizeof(struct tc902x_nic_private),
		            GFP_KERNEL);

	if (sp == NULL)
	{
		printk(KERN_INFO "%s: No memory available for TC9020/9021 private strucutre.\n", tc902x_ethernet_device->name);
		return -ENOMEM;
	}
	else
	{
		/* Fill the allocated memory space with 0s.
		 * Essentially sets all tc902x_nic_private
		 * structure fields to 0.
		 */
		memset(sp, 0, sizeof(*sp));
		tc902x_ethernet_device->priv = sp;
	}

	/* Assign the new device to the list of TC902X Ethernet devices. */
	sp->next_tc902x_ethernet_device = root_tc902x_ethernet_device;
	root_tc902x_ethernet_device = tc902x_ethernet_device;

	/* Declare TC902X NIC functions for Ethernet device methods.
	 */
	tc902x_ethernet_device->open = &tc902x_nic_open;
	tc902x_ethernet_device->stop = &tc902x_nic_stop;
	tc902x_ethernet_device->hard_start_xmit = &tc902x_nic_hard_start_xmit;
	tc902x_ethernet_device->get_stats = &tc902x_nic_get_stats;
	tc902x_ethernet_device->set_multicast_list =
	  &tc902x_nic_set_multicast_list;
	tc902x_ethernet_device->do_ioctl = & tc902x_nic_do_ioctl;
	/* rebuild_header not defined. */
	/* hard_header not defined. */
	/* set_config not defined */
	/* set_mac_address not defined. */
	/* header_cache_bind not defined. */
	/* header_cache_update not defined. */
        tc902x_ethernet_device->change_mtu = &tc902x_nic_change_mtu;
#ifdef TC902X_LINUX2_4
        /* tc902x_ethernet_device->tx_timouet not defined. */
        /* tc902x_ethernet_device->watchdog_timeo not defined. */
#endif

	return 0;
}

int tc902x_nic_do_ioctl(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                        struct ifreq *req, int cmd)
{
	/* IOCTL commands for TC902X NIC.
	 *
	 * SIOCDEVPRIVATE	nothing
	 * SIOCDEVPRIVATE+1	register read
	 *			ifr_data[0] = 0x08, 0x10, 0x20
	 *			ifr_data[1] = register offset
	 *			ifr_data[2] = value read
	 * SIOCDEVPRIVATE+2	register write
	 *			ifr_data[0] = 0x08, 0x10, 0x20
	 *			ifr_data[1] = register offset
	 *			ifr_data[2] = value to write
	 * SIOCDEVPRIVATE+3	GMII register read
	 *			ifr_data[1] = register offset
	 * SIOCDEVPRIVATE+4	GMII register write
	 *			ifr_data[1] = register offset
	 *			ifr_data[2] = value to write
	 * SIOCDEVPRIVATE+5	PCI register read
	 *			ifr_data[0] = 0x08, 0x10, 0x20
	 *			ifr_data[1] = register offset
	 *			ifr_data[2] = value read
	 * SIOCDEVPRIVATE+6	PCI register write
	 *			ifr_data[0] = 0x08, 0x10, 0x20
	 *			ifr_data[1] = register offset
	 *			ifr_data[2] = value to write
	 *
	 */

	u8				val8;
	u16				val16;
	u32				val32;
	unsigned int			*data;
        int                             phyaddr = 0;
	u32				baseaddr;
	struct tc902x_nic_private	*sp = (struct tc902x_nic_private *)
					      tc902x_ethernet_device->priv;

	TC902X_DEBUG_MSG("_nic_do_ioctl\n");

        data = (unsigned int *)&req->ifr_data;
	baseaddr = tc902x_ethernet_device->base_addr;

        switch(cmd)
	{
	        case SIOCDEVPRIVATE:
		return 0;

		case SIOCDEVPRIVATE+1:
		switch(data[0])
		{
			case 0x08:
			data[2] = TC902X_READ_BYTEREG(baseaddr + data[1]);
			return 0;

			case 0x10:
			data[2] = TC902X_READ_WORDREG(baseaddr + data[1]);
			return 0;

			case 0x20:
			data[2] = TC902X_READ_LONGREG(baseaddr + data[1]);
			return 0;

			default:
			data[2] = 0x00;
			return -EINVAL;
		}

		case SIOCDEVPRIVATE+2:
		switch(data[0])
		{
			case 0x08:
			TC902X_WRITE_BYTEREG(baseaddr + data[1], data[2]);
			return 0;

			case 0x10:
			TC902X_WRITE_WORDREG(baseaddr + data[1], data[2]);
			return 0;

			case 0x20:
			TC902X_WRITE_LONGREG(baseaddr + data[1], data[2]);
			return 0;

			default:
			return -EINVAL;
		}

		case SIOCDEVPRIVATE+3:
        	phyaddr = tc902x_find_phyaddr(tc902x_ethernet_device);

		if (phyaddr == -1)
			return -EINVAL;
 
		data[2] = read_phy_register(tc902x_ethernet_device,
		                            phyaddr, data[1]);

		return 0;

		case SIOCDEVPRIVATE+4:
        	phyaddr = tc902x_find_phyaddr(tc902x_ethernet_device);

		if (phyaddr == -1)
			return -EINVAL;
 
		write_phy_register(tc902x_ethernet_device,
		                   phyaddr, data[1], (u16)data[2]);

		return 0;

		case SIOCDEVPRIVATE+5:
		switch(data[0])
		{
			case 0x08:
			pci_read_config_byte(sp->tc902x_pci_device,data[1],
			                     &val8);
			data[2] = (unsigned int)val8;
			return 0;

			case 0x10:
			pci_read_config_word(sp->tc902x_pci_device,data[1],
			                     &val16);
			data[2] = (unsigned int)val16;
			return 0;

			case 0x20:
			pci_read_config_dword(sp->tc902x_pci_device,data[1],
			                     &val32);
			data[2] = (unsigned int)val32;
			return 0;

			default:
			data[2] = 0x00;
			return -EINVAL;
		}

		case SIOCDEVPRIVATE+6:
		switch(data[0])
		{
			case 0x08:
			pci_write_config_byte(sp->tc902x_pci_device,data[1],
			                      (u8)data[2]);
			return 0;

			case 0x10:
			pci_write_config_word(sp->tc902x_pci_device,data[1],
			                      (u16)data[2]);
			return 0;

			case 0x20:
			pci_write_config_dword(sp->tc902x_pci_device,data[1],
			                       (u32)data[2]);
			return 0;

			default:
			return -EINVAL;
		}

		case SIOCSIFMTU:
		{
			return 0;
		}

		default:
		return -EOPNOTSUPP;
        }
}

int tc902x_nic_change_mtu(TC902X_DEVICE_TYPE *tc902x_ethernet_device,
                          int new_mtu)
{
        /* Function to accomodate changes to Maximum Transfer Unit
         * (or MTU) of TC902X NIC. Cannot use default function since
         * the default will not allow for MTU > 1500 bytes.
         */
 
	TC902X_DEBUG_MSG("_nic_change_mtu\n");

        /* Check that the new MTU value is between 68 (14 byte header, 46
         * byte payload, 4 byte FCS) and TC902X_MAX_RXFRAME_SIZE, which
         * corresponds to the MAXFRAMESIZE register in the TC902X.
         */
        if ((new_mtu < 68) || (new_mtu > TC902X_MAX_RXFRAME_SIZE))
        {
                return -EINVAL;
        }
 
        tc902x_ethernet_device->mtu = new_mtu;
 
        return 0;
}

#ifdef TC902X_LINUX2_2
int tc902x_pcibussearch_linux2_2(void)
{
	/* Search for TC902X based devices on the Ethernet bus.
	 * Code specific to the Linux 2.2 kernel.
	 */

	int				error;
	int				i;
	int				foundtc902xnic = 0;
	TC902X_DEVICE_TYPE		*tc902x_ethernet_device;
	struct tc902x_nic_private	*sp;
	struct pci_dev			*tc902x_pci_device;

	TC902X_DEBUG_MSG("_pcibussearch_linux_2_2\n");

	for(i=0; ; i++)
	{
		if (nics_supported[i].vendorid == 0xFFFF)
			break;

		/* Start with the list of all PCI devices. */
		tc902x_pci_device = pci_devices;

		/* Check each entry in the list of all PCI devices. */
		while (tc902x_pci_device)
		{
			if ((tc902x_pci_device->vendor ==
			     nics_supported[i].vendorid) &&
			    (tc902x_pci_device->device ==
			     nics_supported[i].deviceid))
			{
				foundtc902xnic = 1;

				printk(KERN_INFO "%s found.\n",
				       nics_supported[i].NICname);
		                printk(KERN_INFO "Bus %x Slot %x\n",
				       tc902x_pci_device->bus->number,
				       PCI_SLOT(tc902x_pci_device->devfn));

				tc902x_ethernet_device = NULL;

				/* A TC902X based NIC was found on the PCI bus.
				 * Initialize the NIC.
				 */
				error = tc902x_nic_init(tc902x_ethernet_device);
				if (error < 0)
				{
					printk(KERN_INFO "Could not intialize TC9020/9021 based NIC.\n");
					return error;
				}
				else
				{
					printk(KERN_INFO "Ethernet device registered as: %s\n",
					   root_tc902x_ethernet_device->name);
				}

				sp = (struct tc902x_nic_private *)
				     root_tc902x_ethernet_device->priv;

				/* Save the pointer to the PCI device
				 * information.
				 */
				sp->tc902x_pci_device = tc902x_pci_device;
			}

			/* Move onto the next PCI device in the list. */
			tc902x_pci_device = tc902x_pci_device->next;
		}

	}

	return foundtc902xnic;
}
#endif

#ifdef TC902X_LINUX2_4

/* PCI driver structure for Linux 2.4. */
struct  pci_driver      tc902x_pci_driver =
{
	{NULL, NULL},
        TC902X_DRIVER_NAME,
        pci_devices_supported,
        tc902x_pciprobe_linux2_4,
        tc902x_pciremove_linux2_4,
        NULL,
        NULL
};

void tc902x_pciremove_linux2_4(struct pci_dev *tc902x_pci_device_to_remove)
{
	/* Remove function called when a TC902X device is
	 * to be shut down.
	 */

	TC902X_DEVICE_TYPE		*prev_tc902x_ethernet_device = NULL;
	TC902X_DEVICE_TYPE		*tc902x_ethernet_device = NULL;
	TC902X_DEVICE_TYPE		*tc902x_ethernet_device_to_remove =
	                                  NULL;
	struct tc902x_nic_private	*prev_sp = NULL;
	struct tc902x_nic_private	*sp = NULL;
	struct tc902x_nic_private	*sp_to_remove = NULL;

	TC902X_DEBUG_MSG("_pciremove_linux2_4\n");

	tc902x_ethernet_device = root_tc902x_ethernet_device;

	/* Move through list of Ethernet devices looking for
	 * a match.
	 */
	while (tc902x_ethernet_device)
	{
		sp = (struct tc902x_nic_private *)
		               (tc902x_ethernet_device->priv);

		if (sp->tc902x_pci_device == tc902x_pci_device_to_remove)
		{
			/* Save the pointer to the previous Ethernet
			 * device.
			 */
			tc902x_ethernet_device_to_remove =
			  tc902x_ethernet_device;

			sp_to_remove = sp;

			break;
		}

		/* Save the "previous" device in the list. */
		prev_tc902x_ethernet_device = tc902x_ethernet_device;

		/* Retrieve next Ethernet device to be
		 * released.
		 */
		tc902x_ethernet_device = sp->next_tc902x_ethernet_device;
	}

	/* Check if there is a device to remove. */
	if (tc902x_ethernet_device_to_remove == NULL)
	{
		/* There are no Ethernet devices to remove. */
		printk(KERN_INFO "A device remove request does not match with any Ethernet devices.\n");

		return;
	}

	/* Check to see if we are removing the root device in the list. */
	if (root_tc902x_ethernet_device == tc902x_ethernet_device_to_remove)
	{
		/* Change the root Ethernet device to the next device to be
		 * released.
		 */
		root_tc902x_ethernet_device =
		  sp_to_remove->next_tc902x_ethernet_device;
	}
	else if (sp_to_remove->next_tc902x_ethernet_device != NULL)
		/* Check if we need to re-link the list of devices. */
	{
		/* If the "previous" Ethernet device is NULL,
		 * the device is at the head of the list, and
		 * no re-linking is needed.
		 */
		prev_sp = (struct tc902x_nic_private *)
		          (prev_tc902x_ethernet_device->priv);

		prev_sp->next_tc902x_ethernet_device =
				sp_to_remove->next_tc902x_ethernet_device;
	}

	/* Free memory associated with Ethernet device's
	 * private data structure.
	 */
	if (sp_to_remove)
	{
		kfree(sp_to_remove); 
	}

	printk(KERN_INFO "Un-registering Ethernet device %s\n",
	       tc902x_ethernet_device_to_remove->name);

	/* Un-register Ethernet device. */
	unregister_netdev(tc902x_ethernet_device_to_remove);

	/* Free memory associated with Ethernet device. */
	if (tc902x_ethernet_device_to_remove)
	{
		kfree(tc902x_ethernet_device_to_remove);
	}

	return;
}

int tc902x_pciprobe_linux2_4(struct pci_dev *tc902x_pci_device,
                              const struct pci_device_id *id)
{
	/* Probe function called when a TC902X device is found
	 * on the PCI bus.
	 */

	int				error;
	int				i;
	TC902X_DEVICE_TYPE		*tc902x_ethernet_device = NULL;
	struct tc902x_nic_private	*sp;

	TC902X_DEBUG_MSG("_pciprobe_linux2_4\n");
 
	/* Enable TC902X PCI device in Linux system. */
	error = pci_enable_device(tc902x_pci_device);
	if (error < 0)
	{
		return error;
	}

	/* Get the index for the driver description string. */
	i = id->driver_data;

	printk(KERN_INFO "%s found.\n",
	       nics_supported[i].NICname);
        printk(KERN_INFO "Bus %x Slot %x\n",
	       tc902x_pci_device->bus->number,
	       PCI_SLOT(tc902x_pci_device->devfn));

	/* Configure TC902X PCI device within Linux system as
	 * a bus master.
	 */
	pci_set_master(tc902x_pci_device);

	/* Indicate that we can supply 32 bits of address
	 * during PCI bus mastering.
	 */
	if (pci_dma_supported(tc902x_pci_device, 0xFFFFFFFF) < 0)
	{
		printk(KERN_INFO "pci_dma_supported failed.\n");
		return -ENODEV;
	}

	/* A TC902X based NIC was found on the PCI bus.
	 * Initialize the NIC.
	 */
	if ((tc902x_nic_init(tc902x_ethernet_device)) < 0)
	{
		printk(KERN_INFO "Could not intialize TC9020/9021 based NIC.\n");
		return -ENODEV;
	}
	else
	{
		printk(KERN_INFO "Ethernet device registered as: %s\n",
		       root_tc902x_ethernet_device->name);
	}

	sp = (struct tc902x_nic_private *)root_tc902x_ethernet_device->priv;

	/* Save the pointer to the PCI device information. */
	sp->tc902x_pci_device = tc902x_pci_device;

	return 0;
}
#endif

#ifdef MODULE
/* A modularized driver, i.e. not part of mainstream Linux
 * kernel distribution.
 */
int	init_module(void)
{
	/* Initialize the TC902X driver module. */

	int	foundtc902xnic = 0;

	TC902X_DEBUG_MSG("init_module\n");

        printk(KERN_INFO "%s", version);

	/* Define EXPORT_NO_SYMBOLS macro for specifying no
	 * symbols to be exported.
 	 */
	EXPORT_NO_SYMBOLS;

	/* The CONFIG_PCI macro is defined for systems which
	 * have a PCI bus. If the macro is undefined, there is no
	 * PCI bus and the tc902x cannot be utilized.
	 */
#ifdef CONFIG_PCI

	printk(KERN_INFO "TC902X module searching for Ethernet devices on PCI bus...\n");

#ifdef TC902X_LINUX2_2
	foundtc902xnic = tc902x_pcibussearch_linux2_2();
#endif

#ifdef TC902X_LINUX2_4
	if (pci_register_driver(&tc902x_pci_driver) != 0)
	{
		foundtc902xnic = 1;
	}
	else
	{
		foundtc902xnic = 0;
	}
#endif

	if (foundtc902xnic == 0)
	{
		printk(KERN_INFO "TC9020/9021 based Ethernet device not found on PCI bus.\n");
		return -ENODEV;
	}

#else /* PCI not available. */
		/* PCI not supported in system. */
		printk(KERN_INFO "No PCI bus, required by TC9020/9021.\n");
		return -ENODEV;
#endif

	printk(KERN_INFO "TC902X module loaded.\n");

	return 0;
}

void	cleanup_module(void)
{
	/* Clean up system modifications made by TC902X driver
	 * module. 
	 */

#ifdef TC902X_LINUX2_2

	TC902X_DEVICE_TYPE		*next_tc902x_ethernet_device;
	struct tc902x_nic_private	*sp;

	TC902X_DEBUG_MSG("cleanup_module\n");

	/* Unregister all TC902X NICs from the list of Ethernet
	 * devices.
	 */
	while (root_tc902x_ethernet_device)
	{
		sp = (struct tc902x_nic_private *)
		     (root_tc902x_ethernet_device->priv);

		/* Retrieve next Ethernet device to be
		 * released.
		 */
		next_tc902x_ethernet_device = sp->next_tc902x_ethernet_device;

		printk(KERN_INFO "Un-registering Ethernet device %s\n",
		       root_tc902x_ethernet_device->name);

		/* Un-register Ethernet device. */
		unregister_netdev(root_tc902x_ethernet_device);

		/* Free memory associated with Ethernet device's
		 * private data structure.
		 */
		kfree(sp); 

		/* Free memory associated with Ethernet device. */
		kfree(root_tc902x_ethernet_device);

		/* Move to next Ethernet device. */
		root_tc902x_ethernet_device = next_tc902x_ethernet_device;
	}

#endif

#ifdef TC902X_LINUX2_4
	TC902X_DEBUG_MSG("cleanup_module\n");

	pci_unregister_driver(&tc902x_pci_driver);
#endif

	printk(KERN_INFO "TC902X module unloaded.\n");
}
#else /* not MODULE */

int	tc902x_nic_probe(TC902X_DEVICE_TYPE *dev)
{
	int	error;

	TC902X_DEBUG_MSG("_nic_probe\n");

	error = tc902x_nic_init(dev);
	if (error < 0)
		return -ENODEV;
        printk(KERN_INFO "%s", version);

	return 0;
}

#endif

/* end tc902x.c */
