/*
 *
 * tc902x_tune.h
 *
 * Include file with tunable driver parameters for Gigabit Ethernet
 * device driver for Network Interface Cards (NICs) utilizing the
 * Tamarack Microelectronics Inc. TC902X Gigabit or Triple Speed
 * Ethernet Media Access Controller.
 *
 * Craig Rich
 * Sundance Technology, Inc.
 * 1485 Saratoga Avenue
 * Suite 200
 * San Jose, CA 95129
 * 408 873 4117
 * www.sundanceti.com
 * craig_rich@sundanceti.com
 *
 * Rev  Date     Description
 * --------------------------------------------------------------
 * 0.1  3/30/01  New file created from original tc902x.h
 *
 * 0.2  5/22/01  Added PCI_DEVICE_ID_TAMARACK_TC9020_9021_ALT.
 *
 * 0.3  6/20/01  Added TC902X_ADD_IPCHECKSUM_ON_TX,
 *               TC902X_ADD_TCPCHECKSUM_ON_TX, and
 *               TC902X_ADD_UDPCHECKSUM_ON_TX.
 *
 * 0.4  8/11/01  Added comments about avoiding use of TCP/UDP
 *               checksums for silicon revs B3 and earlier.
 *
 * 0.5  10/30/01 Optimized numerous settings.
 */

/* Define PCI vendor and device IDs if not already
 * defined in pci.h.
 */

#ifndef PCI_VENDOR_ID_SUNDANCE
#  define       PCI_VENDOR_ID_SUNDANCE          0x13F0
#endif
#ifndef PCI_DEVICE_ID_SUNDANCE_TC902X
#  define       PCI_DEVICE_ID_SUNDANCE_ST2021   0x2021
#endif
#ifndef PCI_DEVICE_ID_TAMARACK_TC9020_9021_ALT
#  define       PCI_DEVICE_ID_TAMARACK_TC9020_9021_ALT 0x9021
#endif
#ifndef PCI_DEVICE_ID_TAMARACK_TC9020_9021
#  define       PCI_DEVICE_ID_TAMARACK_TC9020_9021 0x1021
#endif
#ifndef PCI_VENDOR_ID_DLINK
#  define       PCI_VENDOR_ID_DLINK             0x1186
#endif
#ifndef PCI_DEVICE_ID_DLINK_1002
#  define       PCI_DEVICE_ID_DLINK_1002        0x4000
#endif

/* Miscellaneous Constants. */
#define   TRUE  1
#define   FALSE 0

/* Assign TC902X_APPEND_FCS_ON_TX > 0 for auto FCS append on TX. */
#define         TC902X_APPEND_FCS_ON_TX         TRUE

/* Assign TC902X_APPEND_FCS_ON_TX > 0 for auto FCS strip on RX. */
#define         TC902X_STRIP_FCS_ON_RX          TRUE

/* Assign TC902X_DROP_ON_RX_ETH_ERRORS > 0 to drop RX frames with
 * Ethernet errors.
 */
#define         TC902X_DROP_ON_RX_ETH_ERRORS    TRUE

/* Assign TC902X_INSERT_MANUAL_VLAN_TAG > 0 to insert VLAN tags manually
 * (via TFC).
 */
#define		TC902X_INSERT_MANUAL_VLAN_TAG   FALSE

/* Assign TC902X_ADD_IPCHECKSUM_ON_TX > 0 for auto IP checksum on TX. */
#define         TC902X_ADD_IPCHECKSUM_ON_TX     FALSE

/* Assign TC902X_ADD_TCPCHECKSUM_ON_TX > 0 for auto TCP checksum on TX.
 * DO NOT USE FOR SILICON REVISIONS B3 AND EARLIER.
 */
#define         TC902X_ADD_TCPCHECKSUM_ON_TX    FALSE

/* Assign TC902X_ADD_UDPCHECKSUM_ON_TX > 0 for auto UDP checksum on TX.
 * DO NOT USE FOR SILICON REVISIONS B3 AND EARLIER.
 */
#define         TC902X_ADD_UDPCHECKSUM_ON_TX    FALSE

/* If inserting VLAN tags manually, assign the TC902X_MANUAL_VLAN_xx
 * constants as desired.
 */
#define		TC902X_MANUAL_VLAN_VID		0xABC
#define		TC902X_MANUAL_VLAN_CFI		0x1
#define		TC902X_MANUAL_VLAN_USERPRIORITY 0x5

#define         TC902X_IO_REG_RANGE		0xFF
#define         TC902X_MEM_REG_RANGE		0x154
#define         TC902X_DRIVER_NAME		"Sundance Technology TC902X Triple-Speed Ethernet"
#define         TC902X_NIC_PHY_ADDRESS          0x01
#define		TC902X_DMALIST_ALIGN_PAD	0x07
#define		TC902X_MULTICAST_HASHTABLE_SIZE	0x40

/* Number of miliseconds to wait after issuing a software reset.
 * 0x05 <= TC902X_AC_RESETWAIT to account for proper 10Mbps operation.
 */
#define         TC902X_AC_RESETWAIT             0x05

/* Number of TC902X_AC_RESETWAIT timeperiods before declaring timeout. */
#define         TC902X_AC_RESET_TIMEOUT         0x0A

/* Minimum number of miliseconds used to toggle MDC clock during
 * MII/GMII register access.
 */
#define         TC902X_PC_PHYCTRLWAIT           0x01

#define		TC902X_TFDLIST_LENGTH		0x100

/* Number of frames between TxDMAComplete interrupt.
 * 0 < TC902X_FRAMESBETWEENTXDMACOMPLETES <= TC902X_TFDLIST_LENGTH
 */
#define		TC902X_FRAMESBETWEENTXDMACOMPLETES 0x1

/* Size of allocated received buffers. Nominally 0x0600.
 * Define larger if expecting jumbo frames.
 */
#define		TC902X_RXFRAG_SIZE		0x0600

/* TC902X_MAX_RXFRAME_SIZE <= TC902X_RXFRAG_SIZE */
#define		TC902X_MAX_RXFRAME_SIZE		0x0600

#define		TC902X_RFDLIST_LENGTH		0x100

/* Maximum number of RFDs to process per interrupt.
 * 1 < TC902X_MAXRFDPROCESS_COUNT < TC902X_RFDLIST_LENGTH
 */
#define		TC902X_MAXRFDPROCESS_COUNT	0x80

/* Minimum margin between last freed RFD, and current RFD.
 * 1 < TC902X_MINUSEDRFDSTOFREE < TC902X_RFDLIST_LENGTH
 */
#define		TC902X_MINUSEDRFDSTOFREE	0x80

/* Specify priority threshhold for a RxDMAPriority interrupt. */
#define		TC902X_PRIORITY_THRESH		0x07

/* Specify the number of receive frames transferred via DMA
 * before a RX interrupt is issued.
 */
#define		TC902X_RXFRAME_COUNT		0x08

/* Specify the maximum amount of time (in 64ns increments) to wait
 * before issuing a RX interrupt if number of frames received
 * is less than TC902X_RXFRAME_COUNT.
 *
 * Value	Time
 * -------------------
 * 0x3D09	~1ms
 * 0x061A	~100us
 * 0x009C	~10us
 * 0x000F	~1us
 */
#define		TC902X_RXDMAWAIT_TIME		0x009C

/* Key register values loaded at driver start up. */

/* TXDMAPollPeriod is specified in 320ns increments.
 *
 * Value	Time
 * ---------------------
 * 0x00-0x01	320ns
 * 0x03		~1us
 * 0x1F		~10us
 * 0xFF		~82us
 */
#define		TC902X_TXDMAPOLLPERIOD_VALUE	0x26
#define		TC902X_TXSTARTTHRESH_VALUE	0x0FFF

/* TxDMAUrgentThresh specifies the minimum amount of
 * data in the transmit FIFO before asserting an
 * urgent transmit DMA request.
 *
 * Value	Min TxFIFO occupied space before urgent TX request
 * ---------------------------------------------------------------
 * 0x00-0x04	128 bytes (1024 bits)
 * 0x27		1248 bytes (~10000 bits)
 * 0x30		1536 bytes (12288 bits)
 * 0xFF		8192 bytes (65535 bits)
 */ 
#define		TC902X_TXDMAURGENTTHRESH_VALUE	0x04

/* TxDMABurstThresh specifies the minimum amount of
 * free space in the transmit FIFO before asserting an
 * transmit DMA request.
 *
 * Value	Min TxFIFO free space before TX request
 * ----------------------------------------------------
 * 0x00-0x08	256 bytes
 * 0x30		1536 bytes
 * 0xFF		8192 bytes
 */ 
#define		TC902X_TXDMABURSTTHRESH_VALUE	0x30

/* RXDMAPollPeriod is specified in 320ns increments.
 *
 * Value	Time
 * ---------------------
 * 0x00-0x01	320ns
 * 0x03		~1us
 * 0x1F		~10us
 * 0xFF		~82us
 */
#define		TC902X_RXDMAPOLLPERIOD_VALUE	0x01
#define		TC902X_RXEARLYTHRESH_VALUE	0x07FF

/* RxDMAUrgentThresh specifies the minimum amount of
 * free space within the receive FIFO before asserting
 * a urgent receive DMA request.
 *
 * Value	Min RxFIFO free space before urgent RX request
 * ---------------------------------------------------------------
 * 0x00-0x04	128 bytes (1024 bits)
 * 0x27		1248 bytes (~10000 bits)
 * 0x30		1536 bytes (12288 bits)
 * 0xFF		8192 bytes (65535 bits)
 */ 
#define		TC902X_RXDMAURGENTTHRESH_VALUE	0x30

/* RxDMABurstThresh specifies the minimum amount of
 * occupied space within the receive FIFO before asserting
 * a receive DMA request.
 *
 * Value	Min TxFIFO free space before TX request
 * ----------------------------------------------------
 * 0x00-0x08	256 bytes
 * 0x30		1536 bytes
 * 0xFF		8192 bytes
 */
#define		TC902X_RXDMABURSTTHRESH_VALUE	0x30

/* FlowOnThresh specifies the maximum amount of occupied
 * space in the receive FIFO before a PAUSE frame with
 * maximum pause time transmitted.
 *
 * Value	Max RxFIFO occupied space before PAUSE
 * ---------------------------------------------------
 * 0x0000	0 bytes
 * 0x0740	29,696 bytes
 * 0x07FF	32,752 bytes
 */
#define		TC902X_FLOWONTHRESH_VALUE	0x0740

/* FlowOffThresh specifies the minimum amount of occupied
 * space in the receive FIFO before a PAUSE frame with
 * zero pause time is transmitted.
 *
 * Value	Max RxFIFO occupied space before PAUSE
 * ---------------------------------------------------
 * 0x0000	0 bytes
 * 0x00BF	3056 bytes
 * 0x07FF	32,752 bytes
 */
#define		TC902X_FLOWOFFTHRESH_VALUE	0x00BF

/* end tc902x_tune.h */
