/*
 * Kernel dbus integration over a netlink socket
 * 
 * Copyright (C) 2003 Red Hat, Inc. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *                                                                                                                
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *                                                                                                              
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Authors:
 *	Arjan van de Ven  <arjanv@redhat.com>
 *
 *  Portions derived from the netlink code in NetKeeper Firewall for Linux
 */

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/socket.h>
#include <linux/skbuff.h>
#include <linux/netlink.h>
#include <linux/string.h>
#include <linux/dbus.h>
#include <net/sock.h>

/* There is one global netlink socket for DBUS to userspace. */
static struct sock *dbus_sock;    

/* Callback for requests to nk_netlink */
static void dbus_netlink_receive(struct sock *sk, int length)
{
	struct sk_buff *skb;

	/* Since we don't currently take any messages from userspace, just drop them all */
	while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		kfree_skb(skb);
	}
}


static int netlink_send_message(int type, char *message, int length)
{
	struct sk_buff *skb;
	char *data_start;
	
	if (!message)
		return -EINVAL;
	
	if (length > PAGE_SIZE)
		return -EINVAL;
	
	skb = alloc_skb(length, GFP_ATOMIC);
	if (!skb) 
		return -ENOMEM;
	data_start = skb_put(skb, length);
	memcpy(data_start, message, length);
	
	return netlink_broadcast(dbus_sock, skb, 0, (1 << type), GFP_ATOMIC);
		
}


void dbus_send_broadcast(int type, char *message, char *fmt, ...)
{
	char *buffer, *c;
	int length;
	int ret;
	if (!message)
		return;
		
	/* Size limits: both the message and the value are limited to 1024 bytes each */
	if (strlen(message) > 1024)
		return;
		
	buffer = (char*) get_zeroed_page(GFP_ATOMIC);
	if (!buffer)
		return;
		
	strcpy(buffer, message);
	length = strlen(message)+1;
	if (fmt) {
		va_list args;        
		c = buffer+length;
	        va_start(args, fmt);
                vsprintf(c, fmt, args);
 		va_end(args);                                
		length += strlen(buffer+length) + 1;
	}
	ret = netlink_send_message(type, buffer, length);
	free_page((unsigned long)buffer);	
}

EXPORT_SYMBOL_GPL(dbus_send_broadcast);


static int dbus_init(void)
{
	dbus_sock = netlink_kernel_create(NETLINK_DBUS, 0, dbus_netlink_receive, THIS_MODULE);
	if (!dbus_sock) {
		printk(KERN_ERR "Kernel dbus agent: unable to create netlink socket; aborting\n");
		return -ENODEV;
	}
	
	return 0;
}

static void dbus_exit(void)
{
	if (dbus_sock) 
		sock_release(dbus_sock->sk_socket);
}

MODULE_DESCRIPTION("Gateway between the kernel and the userspace dbus daemon");
MODULE_AUTHOR("Arjan van de Ven <arjanv@redhat.com>");
MODULE_LICENSE("GPL");

module_init(dbus_init);
module_exit(dbus_exit);
